import copy
import datetime
import json
import logging
import math
import os
import random
import sys
import uuid
from dataclasses import asdict
from pathlib import Path
from typing import IO, Any, Dict, Iterable, List, Optional, Tuple, Union, cast
from warnings import warn

import numpy as np
import pandas as pd
from typeguard import check_type, typechecked

from gantry.exceptions import GantryBatchCreationException, GantryLoggingException
from gantry.logger.constants import CHUNK_SIZE, BatchType, UploadFileType
from gantry.logger.event_builder import (
    _build_feedback_event,
    _build_prediction_and_feedback_events,
    _build_prediction_event,
    _build_prediction_events,
    _create_timestamp_idx,
)
from gantry.logger.stores import APILogStore
from gantry.logger.types import (
    DataLink,
    IngestFromDataConnectorRequest,
    IngestFromDataConnectorResponse,
    Schedule,
)
from gantry.logger.utils import (
    _batch_fail_msg,
    _batch_success_msg,
    _build_batch_iterator,
    _build_data_link_from_file,
    _check_sample_rate,
    _concurrent_upload_multipart_batch,
    _is_not_empty,
    _log_exception,
)
from gantry.serializers import EventEncoder, serializable_value
from gantry.utils import compute_feedback_id

logger = logging.getLogger(__name__)


def _update_tags_with_env(environment: str, tags: Dict[str, str]) -> None:
    if "env" not in tags:
        tags["env"] = environment


def _default_join_key_gen() -> str:
    return str(uuid.uuid4())


def _resolve_join_key(inputs, feedback_id, feedback_keys, join_key: Optional[str]) -> Optional[str]:
    """To keep backwards compatibility"""
    if feedback_id or feedback_keys:
        warn(
            "Deprecated: feedback_id and feedback_keys are going to be removed soon. "
            "Use 'join_key' instead. See https://docs.gantry.io/docs/logging-feedback-actuals"
            " for more information"
        )
        if join_key:
            raise ValueError("Cannot use 'join_key' with 'feedback_keys' or 'feedback_id'")

        if feedback_id:
            if feedback_keys:
                raise GantryLoggingException(
                    "Cannot specify feedback_id and feedback_keys at same time."
                )

            if isinstance(feedback_id, str):
                # Set the feedback_id directly from the given string.
                return feedback_id
            else:
                # Compute feedback_id just from provided feedback_id dictionary
                return compute_feedback_id(feedback_id, list(feedback_id.keys()))
        else:
            return compute_feedback_id(inputs, feedback_keys)

    return join_key


def _resolve_join_keys(
    size: int, inputs, feedback_ids, feedback_keys, join_keys: Optional[List[str]]
) -> Optional[List[str]]:
    """To keep backwards compatibility
    It can assume sizes match.
    """
    if feedback_ids is None and feedback_keys is None and join_keys is None:
        return None

    if feedback_ids or feedback_keys:
        warn(
            "Deprecated: feedback_ids and feedback_keys are going to be removed soon. "
            "Use 'join_keys' instead. See https://docs.gantry.io/docs/logging-feedback-actuals"
            " for more information"
        )
        if join_keys:
            raise ValueError("Cannot use 'join_keys' with 'feedback_keys' or 'feedback_ids'")

        if feedback_ids:
            if feedback_keys:
                raise GantryLoggingException(
                    "Cannot specify feedback_id and feedback_keys at same time."
                )

    join_keys = join_keys or ([None] * size)  # type: ignore
    feedback_ids = feedback_ids or ([None] * size)  # type: ignore
    inputs = inputs or ([None] * size)  # type: ignore
    return [
        _resolve_join_key(inputs[i], feedback_ids[i], feedback_keys, join_keys[i])  # type: ignore
        for i in range(size)
    ]


def _sample_records(size, sample_rate, inputs, outputs, feedbacks, join_keys, timestamps, tags):
    _check_sample_rate(sample_rate)
    if sample_rate < 1:
        idxs = random.sample(range(size), int(sample_rate * size))
        inputs = [x for idx, x in enumerate(inputs) if idx in idxs] if inputs else None
        outputs = [x for idx, x in enumerate(outputs) if idx in idxs] if outputs else None
        feedbacks = [x for idx, x in enumerate(feedbacks) if idx in idxs] if feedbacks else None
        join_keys = [x for idx, x in enumerate(join_keys) if idx in idxs] if join_keys else None
        timestamps = [x for idx, x in enumerate(timestamps) if idx in idxs] if timestamps else None
        tags = [x for idx, x in enumerate(tags) if idx in idxs] if tags else None

    return inputs, outputs, feedbacks, join_keys, timestamps, tags


class Gantry:
    def __init__(
        self,
        log_store: APILogStore,
        environment: str,
        logging_level: str = "INFO",
    ):
        """
        Initializes a new Gantry client to log predictions and feedback.
        Full list of arguments is defined by :class:`gantry.config.ConfigSchema`.

        Args:
            logs_store (BaseLogStore): The log store to use.
            environment (optional, str): Set the value for the 'env' label attached to data
                instrumented by this client. This value can be overridden for each record using
                the 'env' tag (ie `log_record(..., tag={"env": "prod"})` will override the
                environment tag for that specific record).
                WARNING: this parameter has no effect when using `log_file` functionality,
                it only applies to `log_record`/`log_records`.
            logging_level (options: str): Configure logging level for Gantry system.
        """
        self.log_store = log_store
        self.environment = environment
        self.setup_logger(level=logging_level)

    @_log_exception
    @typechecked
    def log_file(
        self,
        application: str,
        filepath: str,
        version: Optional[Union[int, str]] = None,
        timestamp: Optional[str] = None,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        feedbacks: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        feedback_id: Optional[str] = None,
        feedback_keys: Optional[List[str]] = None,
        sep: str = ",",
        join_key: Optional[str] = None,
    ) -> None:
        """
        Ingest a file containing records into Gantry. File MUST contain a header
        line with column names.

        This method provides parameters to customize column name
        mappings to inputs/outputs/feedback/timestamps/etc.

        All specifications are specifications of file column names

        IMPORTANT: SDK validation will be done only at the header level
        (validating inputs/outputs/feedback parametrization is valid).
        Data itself won't be validated due to potential limitations with
        file size.

        Args:
            application (str): Name of the application to ingest the file into.
            filepath (str): path to the file.
            version (optional, Union[int, str]): Version of the function schema to use
                for validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            timestamp (optional, str): by default, timestamp values will be fetched
                from a 'timestamp' named column. Set this parameter to reference a
                different column.
            inputs (optional, list[str]): by default, all columns that start with 'input'
                will be considered inputs. Alternatively, you can provide a list of
                column names using this parameter to be considered inputs.
            outputs (optional, list[str]): by default, all columns that start with 'output'
                will be considered outputs. Alternatively, you can provide a list of
                column names using this parameter to be considered outputs.
            feedbacks (optional, list[str]): by default, all columns that start with 'feedback'
                will be considered feedbacks. Alternatively, you can provide a list of
                column names using this parameter to be considered feedbacks.
            tags (optional, list[str]): by default, all columns that start with 'tags'
                will be considered tags. Alternatively, you can provide a list of
                column names using this parameter to be considered tags.
            feedback_id (optional, str): by default, feedback_id values will be fetched
                from a 'feedback_id' named column. Set this parameter to reference
                another column.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_key' instead.
            feedback_keys (optional, list[str]): provide a list of column names to be used
                as feedback_keys.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_key' instead.
            sep (str, defaults to ','): Character separator, defaults to ',' (for CSV files).
            join_key (optional, str): by default, feedback_id values will be fetched
                from a 'join_key' named column. Set this parameter to reference
                another column.

        """
        data_link = _build_data_link_from_file(
            application=application,
            filepath=filepath,
            sep=sep,
            version=str(version) if isinstance(version, int) else version,
            timestamp=timestamp,
            inputs=inputs,
            outputs=outputs,
            feedback=feedbacks,
            tags=tags,
            feedback_id=feedback_id,
            feedback_keys=feedback_keys,
            join_key=join_key,
        )
        object_size = os.path.getsize(filepath)  # in bytes
        with open(filepath, "r") as f_in:
            self._smart_file_read(data_link, object_size, Path(filepath).name, f_in, CHUNK_SIZE)

    def _smart_file_read(
        self, data_link, object_size, path_name: str, file: IO[str], chunk_size
    ) -> None:
        block_read = lambda: file.read(chunk_size).encode("utf-8")  # noqa: E731
        # Reference on how sentinel parameter works
        # https://docs.python.org/3/library/functions.html#iter
        block_iterator = iter(block_read, b"")
        self._handle_upload(block_iterator, data_link, object_size, f"{uuid.uuid4()}_{path_name}")

    def _handle_upload(
        self, data_batch_iterator: Iterable, data_link: DataLink, object_size: int, filename: str
    ) -> str:
        # Calculate part counts and generate presigned s3 urls.
        total_num_parts = math.ceil(object_size / CHUNK_SIZE)

        logger.info("Initializing upload to Gantry")
        preupload_res = self.log_store._api_client.request(
            "GET",
            "/api/v1/ingest/file-preupload",
            params={
                "filename": filename,
                "num_parts": total_num_parts,
                "filetype": data_link.file_type,
            },
        )
        if "upload_urls" not in preupload_res:
            raise GantryBatchCreationException(
                "Failed to start batch upload. Please check your API key."
            )
        signed_urls = preupload_res["upload_urls"]
        upload_id = preupload_res["upload_id"]
        file_key = preupload_res["key"]

        parts = _concurrent_upload_multipart_batch(data_batch_iterator, signed_urls)

        logger.info("Starting Gantry Ingestion")
        # Submit file completion to gantry API.
        complete_res = self.log_store._api_client.request(
            "POST",
            "/api/v1/ingest/file",
            json={
                "upload_id": upload_id,
                "key": file_key,
                "parts": parts,
                "data_link": asdict(data_link),
            },
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if "batch_id" not in complete_res:
            raise GantryBatchCreationException(
                "Failed to complete upload. Please contact Gantry support."
            )
        batch_id = complete_res["batch_id"]
        _batch_success_msg(batch_id, data_link.application, self.log_store)

        return batch_id

    @_log_exception
    @typechecked
    def log_from_data_connector(
        self,
        application: str,
        source_data_connector_name: str,
        timestamp: Optional[str] = None,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        feedbacks: Optional[List[str]] = None,
        join_key: Optional[str] = None,
        row_tags: Optional[List[str]] = None,
        global_tags: Optional[Dict[str, str]] = None,
        schedule: Optional[Schedule] = None,
        pipeline_name: Optional[str] = None,
    ) -> IngestFromDataConnectorResponse:
        """
        Function to ingest source tabular records from a registered source data
        connector into Gantry.

        To log predictions using this function, both column names of inputs and outputs
        must be passed.
        To log feedback using this function, both column names of join_key and feedback
        must be passed.

        Example:

        .. code-block:: python

            # Record an application's predictions.
            gantry.log_from_data_connector(
                application="foobar",
                source_data_connector_name="my_snowflake_connector",
                inputs=["column_A", "column_B"],
                outputs=["column_C"],
                timestamp="column_T",
                global_tags = {"env":"dev", "version": "1.0.0"},
                row_tags=["column_D"],
                join_key="column_E",
            )

            # Record an application's feedbacks.
            # to a previously ingested prediction.
            gantry.log_from_data_connector(
                application="foobar",
                source_data_connector_name="my_snowflake_connector",
                feedbacks=["column_E", "column_F"],
                timestamp="column_T",
                global_tags = {"env":"dev", "version": "1.0.0"},
                row_tags=["column_D"],
                join_key="column_E",
            )

            # Trigger scheduled ingestion every 8 hours from a data connector incrementally.
            from gantry.logger.types import (
                Schedule,
                ScheduleFrequency,
                ScheduleType,
                ScheduleOptions,
            )

            gantry.log_from_data_connector(
                application="foobar",
                source_data_connector_name="my_snowflake_connector",
                inputs=["column_A", "column_B"],
                outputs=["column_C"],
                timestamp="column_T",
                global_tags = {"env":"dev", "version": "1.0.0"},
                row_tags=["column_D"],
                join_key="column_E",
                schedule=Schedule(
                    start_on="2023-01-14T17:00:00.000000",
                    frequency=ScheduleFrequency.EVERY_8_HOURS,
                    type=ScheduleType.INCREMENTAL_APPEND,
                    options=ScheduleOptions(watermark_key="column_T"),
                )
            )

            # If the data are expected to arrive late in your source table/view, use delay_time
            # to specify how long to wait before triggering the scheduled ingestion.
            gantry.log_from_data_connector(
                application="foobar",
                source_data_connector_name="my_snowflake_connector",
                inputs=["column_A", "column_B"],
                outputs=["column_C"],
                timestamp="column_T",
                global_tags = {"env":"dev", "version": "1.0.0"},
                row_tags=["column_D"],
                join_key="column_E",
                schedule=Schedule(
                    start_on="2023-01-14T17:00:00.000000",
                    frequency=ScheduleFrequency.EVERY_8_HOURS,
                    type=ScheduleType.INCREMENTAL_APPEND,
                    options=ScheduleOptions(
                        delay_time=300, # Delay 5 minutes before triggering ingestion.
                        watermark_key="column_T"
                    ),
                )
            )

        Args:
            application (str): Name of the application to ingest data into
            source_data_connector_name (str): Name of the registered source data connector
                to ingest data from.
            timestamp (optional, str): by default, the timestamp values will be filled
                with the timestamp at ingestion. Set this parameter to reference a
                different column.
            inputs (optional, list[str]): A list of column names for inputs.
            outputs (optional, list[str]): A list ofcolumn names for outputs.
            feedbacks (optional, list[str]): A list of column names for feedbacks.
            join_key (optional, str): A column name to use as the join key to identify each
                record.
            row_tags (optional, list[str]): A list of column names to use the values of
                as tags for each row.
            global_tags (optional, list[dict[str, str]]): A list of dictionaries of string
                key and value pairs to tag the entire ingestion from this data connector
            schedule (optional, Schedule): An optional parameter to schedule the ingestion.
            pipeline_name (optional, str): An optional parameter to specify the ingestion
                pipeline's name
        """
        if schedule and schedule.start_on and isinstance(schedule.start_on, datetime.datetime):
            schedule.start_on = schedule.start_on.isoformat()

        request = IngestFromDataConnectorRequest(
            application=application,
            source_data_connector_name=source_data_connector_name,
            timestamp=timestamp,
            inputs=inputs,
            outputs=outputs,
            feedbacks=feedbacks,
            row_tags=row_tags,
            global_tags=global_tags,
            join_key=join_key,
            schedule=schedule,
            pipeline_name=pipeline_name,
        )

        response = self.log_store.log_from_data_connector_async(request=request)
        logger.info("Successfully submitted ingestion request to Gantry.")

        return response

    @_log_exception
    @typechecked
    def log(
        self,
        application: str,
        version: Optional[Union[int, str]] = None,
        inputs: Optional[Union[dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        outputs: Optional[
            Union[Any, List[Any], dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]
        ] = None,
        feedbacks: Optional[Union[dict, List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        ignore_inputs: Optional[List[str]] = None,
        tags: Optional[Union[Dict[str, str], List[Dict[str, str]]]] = None,
        timestamps: Optional[
            Union[datetime.datetime, List[datetime.datetime], pd.DatetimeIndex, np.ndarray]
        ] = None,
        sort_on_timestamp: bool = True,
        sample_rate: float = 1.0,
        row_tags: Optional[
            Union[Dict[str, str], List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]
        ] = None,
        global_tags: Optional[Dict[str, str]] = None,
        join_keys: Optional[Union[str, List[str], pd.Series]] = None,
        as_batch: Optional[bool] = False,
        run_id: Optional[str] = None,
        run_tags: Optional[dict] = None,
    ):
        """
        A general log function to log inputs, outputs, and feedbacks regardless of
        single or multiple records that is not attached to a specific application.
        This is intended for use primarily in contexts production where contexts
        where avoiding initializing an instance of the application is preferable.

        Args:
            application (str): Name of the application. Gantry validates and monitors data
                by function.
            version (optional, Union[int, str]): Version of the function schema to use for
                validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            inputs (Union[List[dict], pd.Dataframe]): A list of prediction inputs. `inputs[i]`
                is a dict of the features for the i-th prediction to be logged.
            outputs (Union[List[dict], pd.Dataframe]): A list of prediction outputs. `outputs[i]`
                should be the application output for the prediction with features `inputs[i]`.
            feedbacks (Union[List[dict], pd.DataFrame]): A list of feedbacks. `feedbacks[i]`
                is a dict of the features for the i-th prediction to be logged.
            ignore_inputs (optional, List[str]): A list of names of input features that should not
                be monitored.
            timestamps (optional, Union[List[datetime.datetime], pd.DatetimeIndex, np.array[datetime.datetime]):  # noqa: E501
                A list of prediction timestamps.
                If specified, `timestamps[i]` should be the timestamps for the i-th prediction.
                If timestamps = None (default), then the prediction
                timestamp defaults to the time when `log_records` is called.
            sort_on_timestamp (bool, defaults to True): Works when timestamps are provided.
                Sort using the given timestamp. Default to True.
            sample_rate: Used for down-sampling. The probability as a float that each record
                will be sent to Gantry.
            as_batch (bool, defaults to False): Whether to add batch metadata and tracking
                in the 'batch' section of the dashboard
            tags (optional, Optional[Union[Dict[str, str], List[Dict[str, str]]]): A tag is a
                label that you assign to your data. E.g. you can specify which environment
                the data belongs to by setting "env" tag like this tags = {"env": "environment1"}
                if not assigned we will use Gantry client's environment value as the defaults
                environment. If this parameter is a dict, tags will apply to all records.
                Alternatively, you can pass a list of dicts to apply tags to each record
                independantly.
                IMPORTANT: this parameter is in deprecation mode and it will be removed soon.
                Use row_tags and global_tags instead.
            join_keys  (optional, Union[List[str], pd.Series[str]]): provide keys to identify
                each record. These keys can be used later to provide feedback. If not provided,
                a random record key will be generated for each record.
            row_tags (optional, Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]]): Specify row level tags.
                If provided, this parameter should be a list of tags to be applied to each of the records.
                row_tags[i] will contain a dictionary of strings containing the tags to attach to the
                i-th record. Alternatively, tags may be specified by passing in a DataFrame, Series,
                or Array, like inputs.

                For batch global tags, use the 'global_tags' parameter instead.
            global_tags (optional, Dict[str, str]): Specify a set of tags that will be attached to
                all ingested records from this batch. For record specific tags, you can use
                'row_tags' param. Only used when log is not called within a run.
            run_id (optional, str): This should never be provided by user. It will be populated automatically when logging within a run to group records together.
            run_tags (optional, Dict[str, str]): This should never be provided by user. It will be populated automatically when logging within a run to provide global tags for all records in the run.
        """
        if run_id:
            # Generate records/events to be logged as part of the Run.

            # Check if inputs, outputs, feedbacks, timestamps, and join keys are singular.
            # If so, convert to list with one element.
            if isinstance(inputs, dict):
                inputs = [inputs]
            if isinstance(outputs, dict):
                outputs = [outputs]
            if isinstance(feedbacks, dict):
                feedbacks = [feedbacks]
            if isinstance(timestamps, datetime.datetime):
                timestamps = [timestamps]
            if isinstance(join_keys, str):
                join_keys = [join_keys]
            return self.generate_records(
                application=application,
                inputs=inputs,
                outputs=outputs,
                feedbacks=feedbacks,
                global_tags=run_tags,
                row_tags=row_tags,
                timestamps=timestamps,
                join_keys=join_keys,
                run_id=run_id,
                version=version,
                sample_rate=sample_rate,
                sort_on_timestamp=sort_on_timestamp,
                ignore_inputs=ignore_inputs,
            )[0]
        else:
            # Log as a single call.
            if (
                isinstance(inputs, dict)
                and isinstance(outputs, dict)
                and isinstance(feedbacks, dict)
            ):
                return self.single_log_record(
                    application=application,
                    version=version,
                    inputs=inputs,
                    outputs=outputs,
                    feedback=feedbacks,
                    tags=tags,
                    timestamp=timestamps,
                    join_key=join_keys,
                    sample_rate=sample_rate,
                )
            else:
                return self.single_log_records(
                    application=application,
                    version=version,
                    inputs=inputs,
                    outputs=outputs,
                    feedbacks=feedbacks,
                    tags=tags,
                    row_tags=row_tags,
                    global_tags=global_tags,
                    timestamps=timestamps,
                    join_keys=join_keys,
                    as_batch=as_batch,
                    sort_on_timestamp=sort_on_timestamp,
                    sample_rate=sample_rate,
                )

    @_log_exception
    @typechecked
    def generate_records(  # noqa: C901
        self,
        application: str,
        version: Optional[Union[int, str]] = None,
        inputs: Optional[Union[List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        outputs: Optional[Union[List[Any], List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        feedbacks: Optional[Union[List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        sample_rate: float = 1.0,
        run_id: Optional[str] = None,
        tags: Optional[Union[Dict[str, str], List[Dict[str, str]]]] = None,
        join_keys: Optional[Union[List[str], pd.Series]] = None,
        row_tags: Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]] = None,
        global_tags: Optional[Dict[str, str]] = None,
        feedback_keys: Optional[List[str]] = None,
        feedback_ids: Optional[Union[List[str], List[dict]]] = None,
    ):
        """
        Function to generate a batch of events containing predictions (inputs and outputs),
        feedback, or both simultaneously.
        To generate predictions using this method, both inputs and outputs must be passed.
        To generate feedbacks using this method, both join_keys and feedbacks must be passed.

        Args:
            application (str): Name of the application. Gantry validates and monitors data
                by function.
            version (optional, Union[int, str]): Version of the function schema to use for
                validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            inputs (Union[List[dict], pd.Dataframe]): A list of prediction inputs. `inputs[i]`
                is a dict of the features for the i-th prediction to be logged.
            outputs (Union[List[dict], pd.Dataframe]): A list of prediction outputs. `outputs[i]`
                should be the application output for the prediction with features `inputs[i]`.
            feedback_keys (optional, List[str]): A list of names of input features to use for
                feedback lookup. When you later provide a feedback event or label for
                performance metric calculation, you will provide the values of the features
                in this list for Gantry to look up the corresponding prediction.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_keys' instead.
            feedback_ids (optional, List[str] or List[dict]): A list of prediction feedback ids.
                The i-th entry corresponds to the argument `feedback_ids` in
                :meth:`gantry.client.Gantry.log_predictions` for the i-th prediction event.
                If the feedback_id is a List[str], then the exact value of the i-th element in the
                list is used as the feedback join id for the i-th event.
                If the feedback_id is a Dict[str], then the values of the dictionary will be hashed
                to create a feedback join id for the i-th event.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_keys' instead.
            feedbacks (Union[List[dict], pd.DataFrame]): A list of feedbacks. `feedbacks[i]`
                is a dict of the features for the i-th prediction to be logged.
            ignore_inputs (optional, List[str]): A list of names of input features that should not
                be monitored.
            timestamps (optional, Union[List[datetime.datetime], pd.DatetimeIndex, np.array[datetime.datetime]):  # noqa: E501
                A list of prediction timestamps.
                If specified, `timestamps[i]` should be the timestamps for the i-th prediction.
                If timestamps = None (default), then the prediction
                timestamp defaults to the time when `log_records` is called.
            sort_on_timestamp (bool, defaults to True): Works when timestamps are provided.
                Sort using the given timestamp. Default to True.
            sample_rate: Used for down-sampling. The probability as a float that each record
                will be sent to Gantry.
            as_batch (bool, defaults to False): Whether to add batch metadata and tracking
                in the 'batch' section of the dashboard
            tags (optional, Optional[Union[Dict[str, str], List[Dict[str, str]]]): A tag is a
                label that you assign to your data. E.g. you can specify which environment
                the data belongs to by setting "env" tag like this tags = {"env": "environment1"}
                if not assigned we will use Gantry client's environment value as the defaults
                environment. If this parameter is a dict, tags will apply to all records.
                Alternatively, you can pass a list of dicts to apply tags to each record
                independantly.
                IMPORTANT: this parameter is in deprecation mode and it will be removed soon.
                Use row_tags and global_tags instead.
            join_keys  (optional, Union[List[str], pd.Series[str]]): provide keys to identify
                each record. These keys can be used later to provide feedback. If not provided,
                a random record key will be generated for each record.
            row_tags (optional, Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]]): Specify row level tags.
                If provided, this parameter should be a list of tags to be applied to each of the records.
                row_tags[i] will contain a dictionary of strings containing the tags to attach to the
                i-th record. Alternatively, tags may be specified by passing in a DataFrame, Series,
                or Array, like inputs.
                For batch global tags, use the 'global_tags' parameter instead.
            global_tags (optional, Dict[str, str]): Specify a set of tags that will be attached to
                all ingested records from this batch. For record specific tags, you can use
                'row_tags' param.
        Returns:
            Tuple[batch_id, list[join_keys]]: The batch_id will be None if records are not
            logged as batch. The list of join_keys will be the records keys.
        """
        if tags:
            warn(
                "Deprecated: tags parameter is going to be removed soon. "
                "Use row_tags/global_tags instead."
            )

        if inputs is not None and isinstance(inputs, pd.Series):
            inputs = inputs.to_frame()
        if outputs is not None and isinstance(outputs, pd.Series):
            outputs = outputs.to_frame()
        if row_tags is not None and isinstance(row_tags, pd.Series):
            row_tags = row_tags.to_frame()
        if feedbacks is not None and isinstance(feedbacks, pd.Series):
            feedbacks = feedbacks.to_frame()
        if join_keys is not None and isinstance(join_keys, pd.Series):
            join_keys = join_keys.to_list()
            check_type("join_keys", join_keys, List[str])

        if timestamps is not None and isinstance(timestamps, np.ndarray):
            timestamps = timestamps.tolist()
            check_type("timestamps", timestamps, List[datetime.datetime])

        some_preds_exist = _is_not_empty(inputs) or _is_not_empty(outputs)
        preds_exist = _is_not_empty(inputs) and _is_not_empty(outputs)
        feedbacks_exist = _is_not_empty(feedbacks)

        if not preds_exist and some_preds_exist:
            raise ValueError(
                "Tried to log records with incomplete prediction "
                "(both inputs and outputs should be provided)"
            )

        sizes = {}
        if row_tags is not None:
            row_tags = serializable_value(row_tags)
            check_type("row_tags", row_tags, List[Dict[str, str]])

        if feedbacks_exist:
            feedbacks = serializable_value(feedbacks)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("feedbacks", feedbacks, List[dict])
            sizes["feedback"] = len(feedbacks)  # type: ignore
        if preds_exist:
            inputs = serializable_value(inputs)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("inputs", inputs, List[dict])
            sizes["inputs"] = len(inputs)  # type: ignore

            outputs = serializable_value(outputs)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("outputs", outputs, List[dict])
            sizes["outputs"] = len(outputs)  # type: ignore

        if timestamps:
            sizes["timestamps"] = len(timestamps)  # type: ignore
        if join_keys:
            sizes["join_keys"] = len(join_keys)  # type: ignore
        if row_tags:
            sizes["row_tags"] = len(row_tags)  # type: ignore
        if tags and isinstance(tags, List):
            sizes["tags"] = len(tags)  # type: ignore

        unique_sizes = set(sizes.values())
        if len(unique_sizes) != 1:
            raise ValueError(f"Data sizes should match. Got {sizes}")

        size = list(unique_sizes)[0]

        if tags is None or isinstance(tags, Dict):
            tags = [copy.deepcopy(tags) if tags else {} for _ in range(size)]
            tags = cast(List[Dict[str, str]], tags)
        if global_tags:
            for t in tags:
                t.update(global_tags)
        if row_tags:
            for t, row_tag in zip(tags, row_tags):
                t.update(row_tag)

        for t in tags:
            _update_tags_with_env(self.environment, t)
        join_keys = _resolve_join_keys(size, inputs, feedback_ids, feedback_keys, join_keys)
        inputs, outputs, feedbacks, join_keys, timestamps, tags = _sample_records(
            size, sample_rate, inputs, outputs, feedbacks, join_keys, timestamps, tags
        )
        if preds_exist and feedbacks_exist:
            return self._generate_prediction_and_feedback_events(
                application=application,
                inputs=inputs,
                outputs=outputs,
                feedbacks=feedbacks,
                join_keys=join_keys or [_default_join_key_gen() for _ in range(size)],
                version=version,
                ignore_inputs=ignore_inputs,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                run_id=run_id,
                tags=tags,
            )
        elif preds_exist:
            return self._generate_prediction_events(
                application=application,
                inputs=inputs,
                outputs=outputs,
                join_keys=join_keys or [_default_join_key_gen() for _ in range(size)],
                version=version,
                ignore_inputs=ignore_inputs,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                run_id=run_id,
                tags=tags,
            )
        elif feedbacks_exist:
            if join_keys is None:
                raise ValueError(
                    "Can't submit feedback without submitting join_keys "
                    "(or feedback_ids/feedback_keys)"
                )
            return self._generate_feedback_events(
                application=application,
                feedbacks=feedbacks,
                join_keys=join_keys,
                feedback_version=version,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                run_id=run_id,
                tags=tags,
            )
        else:
            raise ValueError("Tried to log_records without prediction and without feedback")

    @_log_exception
    @typechecked
    def single_log_records(  # noqa: C901
        self,
        application: str,
        version: Optional[Union[int, str]] = None,
        inputs: Optional[Union[List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        outputs: Optional[Union[List[Any], List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        feedback_keys: Optional[List[str]] = None,
        feedback_ids: Optional[Union[List[str], List[dict]]] = None,
        feedbacks: Optional[Union[List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        sample_rate: float = 1.0,
        as_batch: bool = False,
        tags: Optional[Union[Dict[str, str], List[Dict[str, str]]]] = None,
        join_keys: Optional[Union[List[str], pd.Series]] = None,
        row_tags: Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]] = None,
        global_tags: Optional[Dict[str, str]] = None,
    ) -> Tuple[Optional[str], List[str]]:
        """
        Function to record a batch of events containing predictions (inputs and outputs),
        feedback, or both simultaneously.
        To log predictions using this method, both inputs and outputs must be passed.
        To log feedbacks using this method, both join_keys and feedbacks must be passed.

        Args:
            application (str): Name of the application. Gantry validates and monitors data
                by function.
            version (optional, Union[int, str]): Version of the function schema to use for
                validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            inputs (Union[List[dict], pd.Dataframe]): A list of prediction inputs. `inputs[i]`
                is a dict of the features for the i-th prediction to be logged.
            outputs (Union[List[dict], pd.Dataframe]): A list of prediction outputs. `outputs[i]`
                should be the application output for the prediction with features `inputs[i]`.
            feedback_keys (optional, List[str]): A list of names of input features to use for
                feedback lookup. When you later provide a feedback event or label for
                performance metric calculation, you will provide the values of the features
                in this list for Gantry to look up the corresponding prediction.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_keys' instead.
            feedback_ids (optional, List[str] or List[dict]): A list of prediction feedback ids.
                The i-th entry corresponds to the argument `feedback_ids` in
                :meth:`gantry.client.Gantry.log_predictions` for the i-th prediction event.
                If the feedback_id is a List[str], then the exact value of the i-th element in the
                list is used as the feedback join id for the i-th event.
                If the feedback_id is a Dict[str], then the values of the dictionary will be hashed
                to create a feedback join id for the i-th event.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_keys' instead.
            feedbacks (Union[List[dict], pd.DataFrame]): A list of feedbacks. `feedbacks[i]`
                is a dict of the features for the i-th prediction to be logged.
            ignore_inputs (optional, List[str]): A list of names of input features that should not
                be monitored.
            timestamps (optional, Union[List[datetime.datetime], pd.DatetimeIndex, np.array[datetime.datetime]):  # noqa: E501
                A list of prediction timestamps.
                If specified, `timestamps[i]` should be the timestamps for the i-th prediction.
                If timestamps = None (default), then the prediction
                timestamp defaults to the time when `log_records` is called.
            sort_on_timestamp (bool, defaults to True): Works when timestamps are provided.
                Sort using the given timestamp. Default to True.
            sample_rate: Used for down-sampling. The probability as a float that each record
                will be sent to Gantry.
            as_batch (bool, defaults to False): Whether to add batch metadata and tracking
                in the 'batch' section of the dashboard
            tags (optional, Optional[Union[Dict[str, str], List[Dict[str, str]]]): A tag is a
                label that you assign to your data. E.g. you can specify which environment
                the data belongs to by setting "env" tag like this tags = {"env": "environment1"}
                if not assigned we will use Gantry client's environment value as the defaults
                environment. If this parameter is a dict, tags will apply to all records.
                Alternatively, you can pass a list of dicts to apply tags to each record
                independantly.
                IMPORTANT: this parameter is in deprecation mode and it will be removed soon.
                Use row_tags and global_tags instead.
            join_keys  (optional, Union[List[str], pd.Series[str]]): provide keys to identify
                each record. These keys can be used later to provide feedback. If not provided,
                a random record key will be generated for each record.
            row_tags (optional, Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]]): Specify row level tags.
                If provided, this parameter should be a list of tags to be applied to each of the records.
                row_tags[i] will contain a dictionary of strings containing the tags to attach to the
                i-th record. Alternatively, tags may be specified by passing in a DataFrame, Series,
                or Array, like inputs.
                For batch global tags, use the 'global_tags' parameter instead.
            global_tags (optional, Dict[str, str]): Specify a set of tags that will be attached to
                all ingested records from this batch. For record specific tags, you can use
                'row_tags' param.
        Returns:
            Tuple[batch_id, list[join_keys]]: The batch_id will be None if records are not
            logged as batch. The list of join_keys will be the records keys.
        """
        if tags:
            warn(
                "Deprecated: tags parameter is going to be removed soon. "
                "Use row_tags/global_tags instead."
            )

        if inputs is not None and isinstance(inputs, pd.Series):
            inputs = inputs.to_frame()
        if outputs is not None and isinstance(outputs, pd.Series):
            outputs = outputs.to_frame()
        if row_tags is not None and isinstance(row_tags, pd.Series):
            row_tags = row_tags.to_frame()
        if feedbacks is not None and isinstance(feedbacks, pd.Series):
            feedbacks = feedbacks.to_frame()
        if join_keys is not None and isinstance(join_keys, pd.Series):
            join_keys = join_keys.to_list()
            check_type("join_keys", join_keys, List[str])

        if timestamps is not None and isinstance(timestamps, np.ndarray):
            timestamps = timestamps.tolist()
            check_type("timestamps", timestamps, List[datetime.datetime])

        some_preds_exist = _is_not_empty(inputs) or _is_not_empty(outputs)
        preds_exist = _is_not_empty(inputs) and _is_not_empty(outputs)
        feedbacks_exist = _is_not_empty(feedbacks)

        if not preds_exist and some_preds_exist:
            raise ValueError(
                "Tried to log records with incomplete prediction "
                "(both inputs and outputs should be provided)"
            )

        sizes = {}
        if row_tags is not None:
            row_tags = serializable_value(row_tags)
            check_type("row_tags", row_tags, List[Dict[str, str]])

        if feedbacks_exist:
            feedbacks = serializable_value(feedbacks)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("feedbacks", feedbacks, List[dict])
            sizes["feedback"] = len(feedbacks)  # type: ignore
        if preds_exist:
            inputs = serializable_value(inputs)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("inputs", inputs, List[dict])
            sizes["inputs"] = len(inputs)  # type: ignore

            outputs = serializable_value(outputs)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("outputs", outputs, List[dict])
            sizes["outputs"] = len(outputs)  # type: ignore

        if timestamps:
            sizes["timestamps"] = len(timestamps)  # type: ignore
        if join_keys:
            sizes["join_keys"] = len(join_keys)  # type: ignore
        if row_tags:
            sizes["row_tags"] = len(row_tags)  # type: ignore
        if tags and isinstance(tags, List):
            sizes["tags"] = len(tags)  # type: ignore

        unique_sizes = set(sizes.values())
        if len(unique_sizes) != 1:
            raise ValueError(f"Data sizes should match. Got {sizes}")

        size = list(unique_sizes)[0]

        if tags is None or isinstance(tags, Dict):
            tags = [copy.deepcopy(tags) if tags else {} for _ in range(size)]
            tags = cast(List[Dict[str, str]], tags)

        if global_tags:
            for t in tags:
                t.update(global_tags)
        if row_tags:
            for t, row_tag in zip(tags, row_tags):
                t.update(row_tag)

        for t in tags:
            _update_tags_with_env(self.environment, t)

        join_keys = _resolve_join_keys(size, inputs, feedback_ids, feedback_keys, join_keys)
        inputs, outputs, feedbacks, join_keys, timestamps, tags = _sample_records(
            size, sample_rate, inputs, outputs, feedbacks, join_keys, timestamps, tags
        )

        if preds_exist and feedbacks_exist:
            return self._log_prediction_and_feedback_events(
                application=application,
                inputs=inputs,
                outputs=outputs,
                feedbacks=feedbacks,
                ignore_inputs=ignore_inputs,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                tags=tags,
                join_keys=join_keys or [_default_join_key_gen() for _ in range(size)],
                version=version,
                as_batch=as_batch,
            )
        elif preds_exist:
            return self._log_prediction_events(
                application=application,
                inputs=inputs,
                outputs=outputs,
                ignore_inputs=ignore_inputs,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                tags=tags,
                join_keys=join_keys or [_default_join_key_gen() for _ in range(size)],
                version=version,
                as_batch=as_batch,
            )
        elif feedbacks_exist:
            if join_keys is None:
                raise ValueError(
                    "Can't submit feedback without submitting join_keys "
                    "(or feedback_ids/feedback_keys)"
                )
            return self._log_feedback_events(
                application=application,
                feedbacks=feedbacks,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                tags=tags,
                join_keys=join_keys,
                feedback_version=version,
                as_batch=as_batch,
            )
        else:
            raise ValueError("Tried to log_records without prediction and without feedback")

    @_log_exception
    @typechecked
    def single_log_record(
        self,
        application: str,
        version: Optional[Union[int, str]] = None,
        inputs: Optional[dict] = None,
        outputs: Optional[Any] = None,
        feedback_id: Optional[Union[str, dict]] = None,
        feedback: Optional[dict] = None,
        feedback_keys: Optional[List[str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamp: Optional[datetime.datetime] = None,
        sample_rate: float = 1.0,
        tags: Optional[Dict[str, str]] = None,
        join_key: Optional[str] = None,
    ) -> Optional[str]:
        """
        Function to record an event, regardless of inputs. Allows client to pass in inputs,
        predictions, and/or feedback for one record.
        Logs error for any individual failures, but does not fail unless method called with
        invalid parameters.
        To log predictions using this method, both inputs and outputs must be passed.
        To log feedback using this method, both join_key and feedback must be passed.

        Args:
            application (str): Name of the application. Gantry validates and monitors data by
                function.
            version (optional, Union[int, str]): Version of the function schema to use
                for validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            inputs (optional, dict): Inputs to the prediction. A dict where the keys are the feature
                names and the values are the feature values.
            outputs (optional, Any): application output on the prediction.
            feedback_id (optional, Dict[str, Any]): A dictionary mapping string keys to values on
                which the feedback id is computed for matching prediction and feedback events.
                Should be the same as the argument `feedback_id` to
                :meth:`gantry.client.Gantry.log_feedback_event` for the matching feedback event.
                DEPRECATION WARNING: this parameter will be removed soon. Use instead 'join_key'.
            feedback (optional, dict): application feedback object.
            feedback_keys (optional, list[str]): A list of names of input features to use for
                feedback lookup. When you later provide a feedback event or label for performance
                metric calculation, you will provide the values of the features in this list for
                Gantry to look up the corresponding prediction.
                DEPRECATION WARNING: this parameter will be removed soon. Use instead 'join_key'.
            ignore_inputs (optional, list[str]): A list of names of input features that should not
                be monitored.
            timestamp (optional, datetime.datetime): Specify a custom timestamp for the when the
                prediction occured. Useful for recording predictions from the past. If not
                specified, then the prediction timestamp defaults to when `log_record` was
                called.
            sample_rate (optional, float): Specify the probability as a float that the event will
                be sent to Gantry.
            tags (optional, Optional[Dict[str, str]]): A tag is a label that you assign to your
                data. E.g. you can specify which environment the data belongs to by setting "env"
                tag like this tags = {"env": "environment1"} if not assigned we will use Gantry
                client's environment value as the default environment.
            join_key  (optional, str): provide a key to identify the record. This key can be used
                later to provide feedback to this record. If not provided, a random record key will
                be generated.
        Returns:
            The record key if record was logged. None in case the sample rate ommited this record.
        """
        # In order to keep backwards compatibility.
        # When feedback_keys and feedback_id get deprecated, this
        # line should just be 'join_key = join_key or _default_join_key_gen()'
        join_key = _resolve_join_key(inputs, feedback_id, feedback_keys, join_key)

        some_pred_exist = _is_not_empty(inputs) or _is_not_empty(outputs)
        pred_exist = _is_not_empty(inputs) and _is_not_empty(outputs)
        feedback_exist = _is_not_empty(feedback)

        tags = tags or {}
        _update_tags_with_env(self.environment, tags)

        if (not pred_exist) and some_pred_exist:
            raise ValueError(
                "Tried to log records with incomplete prediction "
                "(both inputs and outputs should be provided)"
            )

        _check_sample_rate(sample_rate)
        if random.random() > sample_rate:
            return None

        if pred_exist and feedback_exist:
            return self._log_prediction_and_feedback_event(
                application=application,
                version=version,
                inputs=inputs,
                outputs=outputs,
                feedback=feedback,
                join_key=join_key or _default_join_key_gen(),
                ignore_inputs=ignore_inputs,
                timestamp=timestamp,
                tags=tags,
            )
        elif pred_exist:
            return self._log_prediction_event(
                application=application,
                inputs=inputs,
                outputs=outputs,
                join_key=join_key or _default_join_key_gen(),
                version=version,
                ignore_inputs=ignore_inputs,
                timestamp=timestamp,
                tags=tags,
            )
        elif feedback_exist:
            if join_key is None:
                raise ValueError(
                    "Can't submit feedback without submitting join_key "
                    "(or feedback_id/feedback_keys)"
                )

            return self._log_feedback_event(
                application=application,
                join_key=join_key,
                feedback=feedback,
                feedback_version=version,
                timestamp=timestamp,
                tags=tags,
            )
        else:
            logger.error("Tried to log record without prediction and without feedback")

        return None

    @_log_exception
    @typechecked
    def log_records(  # noqa: C901
        self,
        application: str,
        version: Optional[Union[int, str]] = None,
        inputs: Optional[Union[List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        outputs: Optional[Union[List[Any], List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        feedback_keys: Optional[List[str]] = None,
        feedback_ids: Optional[Union[List[str], List[dict]]] = None,
        feedbacks: Optional[Union[List[dict], pd.DataFrame, pd.Series, np.ndarray]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        sample_rate: float = 1.0,
        as_batch: bool = False,
        tags: Optional[Union[Dict[str, str], List[Dict[str, str]]]] = None,
        join_keys: Optional[Union[List[str], pd.Series]] = None,
        row_tags: Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]] = None,
        global_tags: Optional[Dict[str, str]] = None,
    ) -> Tuple[Optional[str], List[str]]:
        """
        Function to record a batch of events containing predictions (inputs and outputs),
        feedback, or both simultaneously.
        To log predictions using this method, both inputs and outputs must be passed.
        To log feedbacks using this method, both join_keys and feedbacks must be passed.

        Example:

        .. code-block:: python

            gantry.log_records(
                application='foobar',
                inputs=[{'A': 100}, {'A': 101}],
                outputs=[{'B': 'foo'}, {'B': 'bar'}],
                version=1,
                feedbacks=[{'B': 'bar'}, {'B': 'foo'}],
                tags={"env": "environment1"},
                join_keys=["12345", "67890"]
            )


        Args:
            application (str): Name of the application. Gantry validates and monitors data
                by function.
            version (optional, Union[int, str]): Version of the function schema to use for
                validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            inputs (Union[List[dict], pd.Dataframe]): A list of prediction inputs. `inputs[i]`
                is a dict of the features for the i-th prediction to be logged.
            outputs (Union[List[dict], pd.Dataframe]): A list of prediction outputs. `outputs[i]`
                should be the application output for the prediction with features `inputs[i]`.
            feedback_keys (optional, List[str]): A list of names of input features to use for
                feedback lookup. When you later provide a feedback event or label for
                performance metric calculation, you will provide the values of the features
                in this list for Gantry to look up the corresponding prediction.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_keys' instead.
            feedback_ids (optional, List[str] or List[dict]): A list of prediction feedback ids.
                The i-th entry corresponds to the argument `feedback_ids` in
                :meth:`gantry.client.Gantry.log_predictions` for the i-th prediction event.
                If the feedback_id is a List[str], then the exact value of the i-th element in the
                list is used as the feedback join id for the i-th event.
                If the feedback_id is a Dict[str], then the values of the dictionary will be hashed
                to create a feedback join id for the i-th event.
                DEPRECATION WARNING: this parameter will be removed soon. Use 'join_keys' instead.
            feedbacks (Union[List[dict], pd.DataFrame]): A list of feedbacks. `feedbacks[i]`
                is a dict of the features for the i-th prediction to be logged.
            ignore_inputs (optional, List[str]): A list of names of input features that should not
                be monitored.
            timestamps (optional, Union[List[datetime.datetime], pd.DatetimeIndex, np.array[datetime.datetime]):  # noqa: E501
                A list of prediction timestamps.
                If specified, `timestamps[i]` should be the timestamps for the i-th prediction.
                If timestamps = None (default), then the prediction
                timestamp defaults to the time when `log_records` is called.
            sort_on_timestamp (bool, defaults to True): Works when timestamps are provided.
                Sort using the given timestamp. Default to True.
            sample_rate: Used for down-sampling. The probability as a float that each record
                will be sent to Gantry.
            as_batch (bool, defaults to False): Whether to add batch metadata and tracking
                in the 'batch' section of the dashboard
            tags (optional, Optional[Union[Dict[str, str], List[Dict[str, str]]]): A tag is a
                label that you assign to your data. E.g. you can specify which environment
                the data belongs to by setting "env" tag like this tags = {"env": "environment1"}
                if not assigned we will use Gantry client's environment value as the defaults
                environment. If this parameter is a dict, tags will apply to all records.
                Alternatively, you can pass a list of dicts to apply tags to each record
                independantly.
                IMPORTANT: this parameter is in deprecation mode and it will be removed soon.
                Use row_tags and global_tags instead.
            join_keys  (optional, Union[List[str], pd.Series[str]]): provide keys to identify
                each record. These keys can be used later to provide feedback. If not provided,
                a random record key will be generated for each record.
            row_tags (optional, Optional[Union[List[Dict[str, str]], pd.DataFrame, pd.Series, np.ndarray]]): Specify row level tags.
                If provided, this parameter should be a list of tags to be applied to each of the records.
                row_tags[i] will contain a dictionary of strings containing the tags to attach to the
                i-th record. Alternatively, tags may be specified by passing in a DataFrame, Series,
                or Array, like inputs.
                For batch global tags, use the 'global_tags' parameter instead.
            global_tags (optional, Dict[str, str]): Specify a set of tags that will be attached to
                all ingested records from this batch. For record specific tags, you can use
                'row_tags' param.
        Returns:
            Tuple[batch_id, list[join_keys]]: The batch_id will be None if records are not
            logged as batch. The list of join_keys will be the records keys.
        """
        if tags:
            warn(
                "Deprecated: tags parameter is going to be removed soon. "
                "Use row_tags/global_tags instead."
            )

        if inputs is not None and isinstance(inputs, pd.Series):
            inputs = inputs.to_frame()
        if outputs is not None and isinstance(outputs, pd.Series):
            outputs = outputs.to_frame()
        if row_tags is not None and isinstance(row_tags, pd.Series):
            row_tags = row_tags.to_frame()
        if feedbacks is not None and isinstance(feedbacks, pd.Series):
            feedbacks = feedbacks.to_frame()
        if join_keys is not None and isinstance(join_keys, pd.Series):
            join_keys = join_keys.to_list()
            check_type("join_keys", join_keys, List[str])

        if timestamps is not None and isinstance(timestamps, np.ndarray):
            timestamps = timestamps.tolist()
            check_type("timestamps", timestamps, List[datetime.datetime])

        some_preds_exist = _is_not_empty(inputs) or _is_not_empty(outputs)
        preds_exist = _is_not_empty(inputs) and _is_not_empty(outputs)
        feedbacks_exist = _is_not_empty(feedbacks)

        if not preds_exist and some_preds_exist:
            raise ValueError(
                "Tried to log records with incomplete prediction "
                "(both inputs and outputs should be provided)"
            )

        sizes = {}
        if row_tags is not None:
            row_tags = serializable_value(row_tags)
            check_type("row_tags", row_tags, List[Dict[str, str]])

        if feedbacks_exist:
            feedbacks = serializable_value(feedbacks)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("feedbacks", feedbacks, List[dict])
            sizes["feedback"] = len(feedbacks)  # type: ignore
        if preds_exist:
            inputs = serializable_value(inputs)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("inputs", inputs, List[dict])
            sizes["inputs"] = len(inputs)  # type: ignore

            outputs = serializable_value(outputs)
            # Check in case this was an np.array
            # Reference https://github.com/gantry-ml/gantry/pull/3341
            check_type("outputs", outputs, List[dict])
            sizes["outputs"] = len(outputs)  # type: ignore

        if timestamps:
            sizes["timestamps"] = len(timestamps)  # type: ignore
        if join_keys:
            sizes["join_keys"] = len(join_keys)  # type: ignore
        if row_tags:
            sizes["row_tags"] = len(row_tags)  # type: ignore
        if tags and isinstance(tags, List):
            sizes["tags"] = len(tags)  # type: ignore

        unique_sizes = set(sizes.values())
        if len(unique_sizes) != 1:
            raise ValueError(f"Data sizes should match. Got {sizes}")

        size = list(unique_sizes)[0]

        if tags is None or isinstance(tags, Dict):
            tags = [copy.deepcopy(tags) if tags else {} for _ in range(size)]
            tags = cast(List[Dict[str, str]], tags)

        if global_tags:
            for t in tags:
                t.update(global_tags)
        if row_tags:
            for t, row_tag in zip(tags, row_tags):
                t.update(row_tag)

        for t in tags:
            _update_tags_with_env(self.environment, t)

        join_keys = _resolve_join_keys(size, inputs, feedback_ids, feedback_keys, join_keys)
        inputs, outputs, feedbacks, join_keys, timestamps, tags = _sample_records(
            size, sample_rate, inputs, outputs, feedbacks, join_keys, timestamps, tags
        )

        if preds_exist and feedbacks_exist:
            return self._log_prediction_and_feedback_events(
                application=application,
                inputs=inputs,
                outputs=outputs,
                feedbacks=feedbacks,
                join_keys=join_keys or [_default_join_key_gen() for _ in range(size)],
                version=version,
                ignore_inputs=ignore_inputs,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                as_batch=as_batch,
                tags=tags,
            )
        elif preds_exist:
            return self._log_prediction_events(
                application=application,
                inputs=inputs,
                outputs=outputs,
                join_keys=join_keys or [_default_join_key_gen() for _ in range(size)],
                version=version,
                ignore_inputs=ignore_inputs,
                timestamps=timestamps,
                sort_on_timestamp=sort_on_timestamp,
                as_batch=as_batch,
                tags=tags,
            )
        elif feedbacks_exist:
            if join_keys is None:
                raise ValueError(
                    "Can't submit feedback without submitting join_keys "
                    "(or feedback_ids/feedback_keys)"
                )

            return self._log_feedback_events(
                application,
                feedbacks,
                join_keys,
                version,
                timestamps,
                sort_on_timestamp,
                as_batch,
                tags,
            )
        else:
            raise ValueError("Tried to log_records without prediction and without feedback")

    @_log_exception
    @typechecked
    def log_record(
        self,
        application: str,
        version: Optional[Union[int, str]] = None,
        inputs: Optional[dict] = None,
        outputs: Optional[Any] = None,
        feedback_id: Optional[Union[str, dict]] = None,
        feedback: Optional[dict] = None,
        feedback_keys: Optional[List[str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamp: Optional[datetime.datetime] = None,
        sample_rate: float = 1.0,
        tags: Optional[Dict[str, str]] = None,
        join_key: Optional[str] = None,
    ) -> Optional[str]:
        """
        Function to record an event, regardless of inputs. Allows client to pass in inputs,
        predictions, and/or feedback for one record.

        Logs error for any individual failures, but does not fail unless method called with
        invalid parameters.

        To log predictions using this method, both inputs and outputs must be passed.
        To log feedback using this method, both join_key and feedback must be passed.

        Example:

        .. code-block:: python

            # Record an application's prediction
            gantry.log_record(
                application='foobar',
                inputs={'A': 100},
                outputs={'prediction': True},
                version=1,
                tags = {"env":"environment1"},
                join_key='12345'
            )

            # Record an application's feedback
            # to a previously ingested prediction.
            gantry.log_record(
                application='foobar',
                feedback={'prediction': False},
                join_key='12345'
            )


        Args:
            application (str): Name of the application. Gantry validates and monitors data by
                function.
            version (optional, Union[int, str]): Version of the function schema to use
                for validation by Gantry.
                If not provided, Gantry will use the latest version. If the version doesn't exist
                yet, Gantry will create it by auto-generating the schema based on data it has seen.
                Providing an int or its value stringified has no difference
                (e.g. version=10 will be the same as version='10').
            inputs (optional, dict): Inputs to the prediction. A dict where the keys are the feature
                names and the values are the feature values.
            outputs (optional, Any): application output on the prediction.
            feedback_id (optional, Dict[str, Any]): A dictionary mapping string keys to values on
                which the feedback id is computed for matching prediction and feedback events.
                Should be the same as the argument `feedback_id` to
                :meth:`gantry.client.Gantry.log_feedback_event` for the matching feedback event.
                DEPRECATION WARNING: this parameter will be removed soon. Use instead 'join_key'.
            feedback (optional, dict): application feedback object.
            feedback_keys (optional, list[str]): A list of names of input features to use for
                feedback lookup. When you later provide a feedback event or label for performance
                metric calculation, you will provide the values of the features in this list for
                Gantry to look up the corresponding prediction.
                DEPRECATION WARNING: this parameter will be removed soon. Use instead 'join_key'.
            ignore_inputs (optional, list[str]): A list of names of input features that should not
                be monitored.
            timestamp (optional, datetime.datetime): Specify a custom timestamp for the when the
                prediction occured. Useful for recording predictions from the past. If not
                specified, then the prediction timestamp defaults to when `log_record` was
                called.
            sample_rate (optional, float): Specify the probability as a float that the event will
                be sent to Gantry.
            tags (optional, Optional[Dict[str, str]]): A tag is a label that you assign to your
                data. E.g. you can specify which environment the data belongs to by setting "env"
                tag like this tags = {"env": "environment1"} if not assigned we will use Gantry
                client's environment value as the default environment.
            join_key  (optional, str): provide a key to identify the record. This key can be used
                later to provide feedback to this record. If not provided, a random record key will
                be generated.

        Returns:
            The record key if record was logged. None in case the sample rate ommited this record.

        """
        warn("Deprecated: log_record is going to be removed soon. " "Use log() instead.")
        # In order to keep backwards compatibility.
        # When feedback_keys and feedback_id get deprecated, this
        # line should just be 'join_key = join_key or _default_join_key_gen()'
        join_key = _resolve_join_key(inputs, feedback_id, feedback_keys, join_key)

        some_pred_exist = _is_not_empty(inputs) or _is_not_empty(outputs)
        pred_exist = _is_not_empty(inputs) and _is_not_empty(outputs)
        feedback_exist = _is_not_empty(feedback)

        tags = tags or {}
        _update_tags_with_env(self.environment, tags)

        if (not pred_exist) and some_pred_exist:
            raise ValueError(
                "Tried to log records with incomplete prediction "
                "(both inputs and outputs should be provided)"
            )

        _check_sample_rate(sample_rate)
        if random.random() > sample_rate:
            return None

        if pred_exist and feedback_exist:
            return self._log_prediction_and_feedback_event(
                application=application,
                version=version,
                inputs=inputs,
                outputs=outputs,
                feedback=feedback,
                join_key=join_key or _default_join_key_gen(),
                ignore_inputs=ignore_inputs,
                timestamp=timestamp,
                tags=tags,
            )
        elif pred_exist:
            return self._log_prediction_event(
                application=application,
                inputs=inputs,
                outputs=outputs,
                join_key=join_key or _default_join_key_gen(),
                version=version,
                ignore_inputs=ignore_inputs,
                timestamp=timestamp,
                tags=tags,
            )
        elif feedback_exist:
            if join_key is None:
                raise ValueError(
                    "Can't submit feedback without submitting join_key "
                    "(or feedback_id/feedback_keys)"
                )

            return self._log_feedback_event(
                application=application,
                join_key=join_key,
                feedback=feedback,
                feedback_version=version,
                timestamp=timestamp,
                tags=tags,
            )
        else:
            logger.error("Tried to log record without prediction and without feedback")

        return None

    @_log_exception
    def ping(self) -> bool:
        """
        Pings the log store API server to check if it is alive.
        Returns True if alive, False if there is an error during ping process.
        """
        return self.log_store.ping()

    @_log_exception
    def ready(self) -> bool:
        """
        Checks if the configured API key authenticates with the API.
        Returns True if ready, False otherwise.
        """
        return self.log_store.ready()

    @_log_exception
    def _log_prediction_and_feedback_events(
        self,
        application: str,
        inputs: List[dict],
        outputs: List[dict],
        feedbacks: List[dict],
        join_keys: List[str],
        version: Optional[Union[int, str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        as_batch: bool = False,
        tags: Optional[List[Dict[str, str]]] = None,
    ) -> Tuple[Optional[str], List[str]]:
        """Internal method to log batch of predictions AND feedbacks
        This method can assume inputs/outputs/feedbacks/timestamps/tags have same size
        """
        size = len(inputs)

        timestamp_idx = _create_timestamp_idx(sort_on_timestamp, timestamps, size)

        events = _build_prediction_and_feedback_events(
            application=application,
            timestamp_idx=timestamp_idx,
            tags=tags,
            version=version,
            inputs=inputs,
            outputs=outputs,
            feedbacks=feedbacks,
            join_keys=join_keys,
            ignore_inputs=ignore_inputs,
        )

        if events:
            if as_batch:
                return (
                    self._upload_data_as_batch(application, version, events, BatchType.RECORD),
                    join_keys,
                )
            else:
                self.log_store.log_batch(application, events)
        else:
            logger.info("No events to log")

        return None, join_keys

    @_log_exception
    def _generate_prediction_and_feedback_events(
        self,
        application: str,
        inputs: List[dict],
        outputs: List[dict],
        feedbacks: List[dict],
        join_keys: List[str],
        version: Optional[Union[int, str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        run_id: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
    ):
        """Internal method to generate batch of predictions AND feedbacks as list of events.
        This method can assume inputs/outputs/feedbacks/timestamps/tags have same size
        """
        size = len(inputs)

        timestamp_idx = _create_timestamp_idx(sort_on_timestamp, timestamps, size)

        events = _build_prediction_and_feedback_events(
            application=application,
            timestamp_idx=timestamp_idx,
            tags=tags,
            version=version,
            inputs=inputs,
            outputs=outputs,
            feedbacks=feedbacks,
            join_keys=join_keys,
            ignore_inputs=ignore_inputs,
            run_id=run_id,
        )

        return events, join_keys

    @_log_exception
    def _log_prediction_and_feedback_event(
        self,
        application: str,
        inputs: dict,
        outputs: Any,
        feedback: Optional[dict],
        join_key: str,
        version: Optional[Union[int, str]] = None,
        feedback_keys: Optional[List[str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamp: Optional[datetime.datetime] = None,
        tags: Optional[Dict[str, str]] = None,
    ) -> str:
        """Internal method to log an event with prediction AND feedback"""
        ev = {}

        ev.update(
            _build_prediction_event(
                inputs,  # type: ignore
                outputs,
                application,
                join_key,
                version,
                ignore_inputs,
                custom_timestamp=timestamp,
                tags=tags,
            )
        )

        fev = _build_feedback_event(
            application,
            join_key,
            feedback,  # type: ignore
            version,
            timestamp,
            tags=tags,
        )

        ev["metadata"].update(fev.pop("metadata"))  # update nested dict separately
        ev.update(fev)
        self.log_store.log(application, ev)
        return join_key

    @_log_exception
    def _log_feedback_event(
        self,
        application: str,
        join_key: str,
        feedback: dict,
        feedback_version: Optional[Union[str, int]] = None,
        timestamp: Optional[datetime.datetime] = None,
        tags: Optional[Dict[str, str]] = None,
    ) -> str:
        """Internal method to log a feedback event"""
        ev = _build_feedback_event(
            application,
            join_key,
            feedback,
            feedback_version,
            timestamp,
            tags=tags,
        )
        self.log_store.log(application, ev)
        return join_key

    @_log_exception
    def _generate_feedback_events(
        self,
        application: str,
        feedbacks: List[dict],
        join_keys: List[str],
        feedback_version: Optional[Union[str, int]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        run_id: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
    ):
        """Internal method to generate batch of feedback events
        This method can assume feedbacks/join_keys/timestamps have same size
        """
        timestamp_idx = _create_timestamp_idx(sort_on_timestamp, timestamps, len(feedbacks))

        if len(join_keys) == len(feedbacks):
            events = []
            for idx, timestamp in timestamp_idx:
                try:
                    events.append(
                        _build_feedback_event(
                            application,
                            join_keys[idx],
                            feedbacks[idx],
                            feedback_version,
                            timestamp,
                            batch_id=run_id,
                            tags=tags[idx] if tags else None,
                        )
                    )
                except GantryLoggingException as le:
                    # this is caused by a user error
                    # log the error without the stacktrace
                    logger.error("Error generating logging data to Gantry: %s", le)
                except Exception as e:
                    logger.error(
                        "Failed to generate feedback with id {} due to {}".format(join_keys[idx], e)
                    )
        else:
            raise GantryLoggingException("Feedback_ids and feedbacks lists don't have same length.")

        return events, join_keys

    @_log_exception
    def _log_feedback_events(
        self,
        application: str,
        feedbacks: List[dict],
        join_keys: List[str],
        feedback_version: Optional[Union[str, int]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        as_batch: bool = False,
        tags: Optional[List[Dict[str, str]]] = None,
    ) -> Tuple[Optional[str], List[str]]:
        """Internal method to log batch of feedback events
        This method can assume feedbacks/join_keys/timestamps have same size
        """
        batch_id = None
        try:
            timestamp_idx = _create_timestamp_idx(sort_on_timestamp, timestamps, len(feedbacks))

            if len(join_keys) == len(feedbacks):
                events = []
                for idx, timestamp in timestamp_idx:
                    try:
                        events.append(
                            _build_feedback_event(
                                application,
                                join_keys[idx],
                                feedbacks[idx],
                                feedback_version,
                                timestamp,
                                batch_id=batch_id,
                                tags=tags[idx] if tags else None,
                            )
                        )
                    except GantryLoggingException as le:
                        # this is caused by a user error
                        # log the error without the stacktrace
                        logger.error("Error logging data to Gantry: %s", le)
                    except Exception as e:
                        logger.error(
                            "Failed to log feedback with id {} due to {}".format(join_keys[idx], e)
                        )

                if events:
                    if as_batch:
                        return (
                            self._upload_data_as_batch(
                                application,
                                feedback_version,
                                events,
                                BatchType.FEEDBACK,
                            ),
                            join_keys,
                        )
                    else:
                        batch_id = None
                        self.log_store.log_batch(application, events)
                        _batch_success_msg(batch_id, application, self.log_store)
                else:
                    raise ValueError("No events to log")
            else:
                raise GantryLoggingException(
                    "Feedback_ids and feedbacks lists don't have same length."
                )
        except Exception as e:
            _batch_fail_msg(batch_id)

            raise e

        return None, join_keys

    @_log_exception
    def _log_prediction_event(
        self,
        application: str,
        inputs: dict,
        outputs: Any,
        join_key: str,
        version: Optional[Union[int, str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamp: Optional[datetime.datetime] = None,
        tags: Optional[Dict[str, str]] = None,
    ) -> str:
        """Internal method to log a prediction event"""
        ev = _build_prediction_event(
            inputs,
            outputs,
            application,
            join_key,
            version,
            ignore_inputs,
            custom_timestamp=timestamp,
            tags=tags,
        )
        self.log_store.log(application, ev)
        return join_key

    @_log_exception
    def _generate_prediction_events(
        self,
        application: str,
        inputs: List[dict],
        outputs: List[dict],
        join_keys: List[str],
        version: Optional[Union[int, str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        run_id: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
    ):
        """
        Internal method to generate a batch of prediction events.
        Assume inputs/outputs/join_keys/timestamps have same size
        """
        timestamp_idx = _create_timestamp_idx(sort_on_timestamp, timestamps, len(inputs))
        events = _build_prediction_events(
            application=application,
            inputs=inputs,
            outputs=outputs,
            timestamp_idx=timestamp_idx,
            tags=tags,
            version=version,
            join_keys=join_keys,
            ignore_inputs=ignore_inputs,
            batch_id=run_id,
        )
        return events, join_keys

    @_log_exception
    def _log_prediction_events(
        self,
        application: str,
        inputs: List[dict],
        outputs: List[dict],
        join_keys: List[str],
        version: Optional[Union[int, str]] = None,
        ignore_inputs: Optional[List[str]] = None,
        timestamps: Optional[Union[List[datetime.datetime], pd.DatetimeIndex, np.ndarray]] = None,
        sort_on_timestamp: bool = True,
        as_batch: bool = False,
        tags: Optional[List[Dict[str, str]]] = None,
    ) -> Tuple[Optional[str], List[str]]:
        """Internal method to log a batch of prediction events
        This method can assume inputs/outputs/join_keys/timestamps have same size
        """
        batch_id = None
        try:
            timestamp_idx = _create_timestamp_idx(sort_on_timestamp, timestamps, len(inputs))

            events = _build_prediction_events(
                application=application,
                inputs=inputs,
                outputs=outputs,
                timestamp_idx=timestamp_idx,
                tags=tags,
                version=version,
                join_keys=join_keys,
                ignore_inputs=ignore_inputs,
                batch_id=batch_id,
            )

            if events:
                if as_batch:
                    return (
                        self._upload_data_as_batch(
                            application, version, events, BatchType.PREDICTION
                        ),
                        join_keys,
                    )
                else:
                    batch_id = None
                    self.log_store.log_batch(application, events)
                    _batch_success_msg(batch_id, application, self.log_store)
            else:
                logger.info("No events to log")
        except Exception as e:
            _batch_fail_msg(batch_id)
            raise e

        return None, join_keys

    def _upload_data_as_batch(
        self,
        application: str,
        version: Optional[Union[str, int]],
        events: List,
        batch_type: BatchType,
        run_id: Optional[str] = None,
        run_name: Optional[str] = None,
        run_tags: Optional[Dict] = None,
    ) -> str:
        data_link = DataLink(
            file_type=UploadFileType.EVENTS,
            batch_type=batch_type,
            num_events=len(events),
            application=application,
            version=str(version) if isinstance(version, int) else version,
            log_timestamp=datetime.datetime.utcnow().isoformat(),
            batch_id=run_id,
            run_name=run_name,
            run_tags=run_tags,
        )
        events_bytesize = sys.getsizeof(json.dumps(events, cls=EventEncoder).encode("utf-8"))
        batch_count = int(
            CHUNK_SIZE / (events_bytesize / data_link.num_events)
        )  # max file chunk size / average event size -> count of batch events to get to chunk.
        return self._handle_upload(
            _build_batch_iterator(events, batch_count),
            data_link,
            events_bytesize,
            f"{application}_{uuid.uuid4()}",
        )

    @staticmethod
    def setup_logger(level="INFO"):
        if not level:
            return

        pkg_logger = logging.getLogger("gantry")
        pkg_logger.setLevel(level)

        existing_handlers = pkg_logger.handlers
        for handler in existing_handlers:
            if isinstance(handler, logging.StreamHandler):
                return
        formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        pkg_logger.addHandler(handler)



import datetime
import json
import logging
import os
import pathlib
import subprocess
import tempfile
import zipfile
from time import sleep
from typing import Dict, List, Tuple

import click
import requests
import yaml
from click_aliases import ClickAliasedGroup

import gantry
from gantry.api_client import APIClient

logger = logging.getLogger(__name__)


PENDING_TASK_STATUSES = [
    "NOT_YET_STARTED",
    "BUILD_IN_PROGRESS",
    "BUILD_SUCCEEDED",
    "PUSH_SUCCEEDED",
    "PUBLISH_LAMBDA_VERSION_SUCCEEDED",
]


@click.group(cls=ClickAliasedGroup)
def projection():
    """
    Use this to register a new custom projection to Gantry.
    """


class DependenciesInstallError(Exception):
    pass


def _update_dev_projection(
    api_client: APIClient, projection: Dict, projection_dir: str, dry_run: bool
):
    """
    For Gantry's developers only: Upload a custom projection directly from a folder
    that contains:
        1. config.yaml: contains custom projection definition
        2. function.zip: source code with all required dependencies to get invoked as a lambda
    """
    click.secho("Parsing custom projection configs...", nl=False, fg="cyan")
    projection = _get_custom_projection_configs(projection_dir)
    click.secho("SUCCESS", fg="cyan")

    click.secho("Validate custom projection definition...", nl=False, fg="cyan")
    valid_config = _validate_custom_projection_configs(api_client, projection)
    if not valid_config:
        click.secho("FAILED.", fg="red")
        raise click.ClickException("Invalid projection definition")
    click.secho("SUCCESS", fg="cyan")

    if dry_run:
        click.secho("Dry run complete.", fg="cyan")
        return True

    click.secho("Set up upload location...", nl=False, fg="cyan")
    retrieved, resp = _get_s3_destination(api_client, projection)
    if not retrieved:
        click.secho("FAILED.", fg="red")
        err_msg = resp.get("error", "")
        raise click.ClickException(f"Fail to set up upload location: {err_msg}")
    click.secho("SUCCESS", fg="cyan")

    upload_url, s3_key = resp["upload_url"], resp["s3_key"]
    zip_filepath = f"{projection_dir}/function.zip"
    proc = subprocess.run(["curl", "--upload-file", zip_filepath, upload_url])
    if proc.returncode != 0:
        return click.ClickException("Failed to upload custom projection function")

    request_sent, task_id, err_msg = _async_update_custom_projection(api_client, s3_key, projection)
    if not request_sent:
        click.secho("FAILED.", fg="red")
        raise click.ClickException(
            f"Request to submit from-file custom projection failed: {err_msg}"
        )
    click.secho(f"SUCCESS. Task {task_id} has been submitted.", fg="cyan")


@projection.command(aliases=["create"], help="Create or update a custom projection in Gantry")
@click.option("--projection-dir", type=click.Path())
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Enable dry run to verify inputs before actually submitting a custom projection to Gantry",
)
def update(projection_dir, dry_run):
    gantry.init()
    api_client = gantry.get_client().log_store._api_client

    click.secho("Parsing custom projection configs...", nl=False, fg="cyan")
    projection = _get_custom_projection_configs(projection_dir)
    click.secho("SUCCESS", fg="cyan")

    if _is_dev_projection(projection):
        _update_dev_projection(api_client, projection, projection_dir, dry_run)
    else:
        _update_projection(api_client, projection, projection_dir, dry_run)


def _update_projection(api_client: APIClient, projection: Dict, projection_dir: str, dry_run: bool):
    click.secho("Validate custom projection definition...", nl=False, fg="cyan")
    valid_config = _validate_custom_projection_configs(api_client, projection)
    if not valid_config:
        click.secho("FAILED.", fg="red")
        raise click.ClickException("Invalid projection definition")
    click.secho("SUCCESS", fg="cyan")

    if dry_run:
        click.secho("Dry run complete.", fg="cyan")
        return True

    click.secho("Set up upload location...", nl=False, fg="cyan")
    retrieved, resp = _get_s3_destination(api_client, projection)
    if not retrieved:
        click.secho("FAILED.", fg="red")
        err_msg = resp.get("error", "")
        raise click.ClickException(f"Fail to set up upload location: {err_msg}")
    click.secho("SUCCESS", fg="cyan")

    upload_url, s3_key = resp["upload_url"], resp["s3_key"]

    click.secho("Upload new definition...", nl=False, fg="cyan")
    with tempfile.TemporaryDirectory() as tmpdirname:
        uploaded, err_msg = _upload_function_zip(
            os.path.join(tmpdirname, "function.zip"), projection_dir, upload_url
        )

    if not uploaded:
        click.secho("FAILED.", fg="red")
        raise click.ClickException(f"Fail to upload new projection definition: {err_msg}")
    click.secho("SUCCESS", fg="cyan")

    click.secho("Submit build custom projection request...", nl=False, fg="cyan")
    request_sent, task_id, err_msg = _async_update_custom_projection(api_client, s3_key, projection)
    if not request_sent:
        click.secho("FAILED.", fg="red")
        raise click.ClickException(f"Request to build custom projection failed: {err_msg}")
    click.secho("SUCCESS", fg="cyan")
    click.secho(f"--> Task {task_id} has been submitted.")

    click.secho("Check build progress...", nl=False, fg="cyan")

    _call_get_build_logs(api_client, task_id)


@projection.command("get-build-logs", help="Display custom projection build logs given task ID")
@click.option("--task-id", type=click.STRING)
def get_build_logs(task_id):
    gantry.init()
    api_client = gantry.get_client().log_store._api_client
    _call_get_build_logs(api_client, task_id)


def _call_get_build_logs(api_client: APIClient, task_id: str):
    status = "NOT_YET_STARTED"
    i, max_attempts, interval_seconds = 0, 60, 10
    while i < max_attempts and status in PENDING_TASK_STATUSES:
        new_status, logs = _get_build_logs(api_client, task_id)
        if status == "ERROR":
            click.secho(f"--> {status}", nl=False, fg="red")
            click.secho(logs)
            break
        elif status == new_status:
            click.secho(".", nl=False)
        else:
            status = new_status
            click.secho("")
            if "FAILED" in status:
                click.secho(f"--> {status}", nl=False, fg="red")
            else:
                click.secho(f"--> {status}", nl=False)
            if logs:
                click.secho("")
                click.secho("Build logs:", fg="cyan")
                click.secho(logs)

        i += 1
        sleep(interval_seconds)

    if i == max_attempts:
        click.secho("Closing connection, custom projection build is still in progress.")


@projection.command("get-logs", help="Retrieve custom projection run logs given a time range")
@click.option("--application-name", type=click.STRING)
@click.option("--projection-name", type=click.STRING)
@click.option(
    "--start",
    type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S"]),
    default=None,
    help="Provide a start time to search for logs",
)
@click.option(
    "--end",
    type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S"]),
    default=None,
    help="Provide an end time to search for logs",
)
@click.option(
    "--tail",
    type=click.INT,
    default=None,
    help="Get the most recent events. If this option is used, --start and --end will be ignored",
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="List raw event logs")
def get_logs(application_name, projection_name, start, end, tail, verbose):
    gantry.init()
    api_client = gantry.get_client().log_store._api_client

    params = {"limit": tail, "start_time": start, "end_time": end}
    response = api_client.request(
        "GET",
        f"/api/v1/{application_name}/custom_projections/{projection_name}/cloudwatch_logs",
        params=params,
    )
    if response.get("response") != "ok":
        err_msg = response.get("error", "unknown error")
        raise click.ClickException(f"Fail to retrieve logs: {err_msg}.")

    event_logs = response.get("logs", [])
    _display_logs(event_logs, verbose)


def _display_logs(events: List[Dict], verbose: bool = False):
    if verbose:
        for event in events:
            click.echo(event)
    else:
        click.secho(f"Found {len(events)} event(s).", fg="cyan")
        for event in events:
            _display_formatted_log(event)


def _display_formatted_log(event: Dict):
    def format_ts(timestamp: int) -> str:
        timestamp_seconds = timestamp / 1000.0
        return datetime.datetime.fromtimestamp(timestamp_seconds).strftime("%Y-%m-%d %H:%M:%S")

    click.secho(f'{format_ts(event["timestamp"])}: ', nl=False, fg="cyan")
    click.secho(event["message"].strip())


def _upload_function_zip(
    zip_file: str, projection_dir: str, presigned_url: str
) -> Tuple[bool, str]:
    zipped, err_msg = _zip_function(projection_dir, zip_file)
    if not zipped:
        return False, err_msg

    resp = requests.put(presigned_url, files={"file": open(zip_file, "rb")})

    if resp.status_code != 200:
        click.secho(resp.text)
        return False, "Failed to upload custom projection function"
    return True, ""


def _zip_function(projection_dir: str, zip_file: str) -> Tuple[bool, str]:
    with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(projection_dir):
            for file in files:
                filepath = os.path.join(root, file)

                relpath = os.path.relpath(filepath, projection_dir)

                # if this file is part of the extra_deps directory
                # move it up and out of the extra_deps directory
                # by removing the first part of the path
                relpath_path = pathlib.Path(relpath)
                if relpath_path.parts[0] == "extra_deps":
                    relpath = str(pathlib.Path(*relpath_path.parts[1:]))

                zipf.write(filepath, relpath)

    return True, ""


def _get_custom_projection_configs(projection_dir: str) -> Dict:
    config_file = "config.yaml"
    projection = {}
    try:
        with open(f"{projection_dir}/{config_file}", "r") as stream:
            projection = yaml.safe_load(stream)
    except FileNotFoundError:
        raise ValueError(f"{config_file} not found under projection directory.")

    return projection


def _validate_custom_projection_configs(api_client: APIClient, projection: Dict) -> bool:
    response = api_client.request(
        "POST",
        "/api/v1/custom_projections/validate",
        data={"projection": json.dumps(projection)},
    )

    if response.get("response") == "ok":
        return True
    if "errors" in response:
        # print each error on a new line
        for err in response["errors"]:
            click.secho(err)
    else:
        if "error" in response:
            error_dict = response["error"]
            if error_dict.get("message", ""):
                click.secho(error_dict.get("message"))
            if error_dict.get("validation_errors", []):
                for err in error_dict["validation_errors"]:
                    click.secho(f"-->{err}")

    return False


def _get_s3_destination(api_client: APIClient, projection: Dict) -> Tuple[bool, Dict]:
    response = api_client.request(
        "GET",
        "/api/v1/custom_projections/pre-upload",
        params={
            "projection_name": projection["function_definition"]["projection_name"],
        },
    )

    if response.get("response") == "ok":
        return True, response
    return False, response


def _async_update_custom_projection(
    api_client: APIClient, s3_key: str, projection: Dict
) -> Tuple[bool, str, str]:
    response = api_client.request(
        "POST",
        "/api/v1/custom_projections",
        data={"s3_key": s3_key, "projection": json.dumps(projection)},
    )

    if response.get("response") == "ok":
        return True, response["task_id"], ""
    return False, "", response.get("error", "Error updating custom projection")


def _get_build_logs(api_client: APIClient, task_id: str) -> Tuple[str, str]:
    response = api_client.request(
        "GET",
        f"/api/v1/custom_projections/logs/{task_id}",
    )
    if response.get("response") == "ok":
        return str(response["status"]), response["logs"]
    return "ERROR", "Error getting task logs"


def _is_dev_projection(projection: Dict) -> bool:
    function_def = projection.get("function_definition")
    if function_def and function_def.get("mode") == "lambda_dev":
        return True
    return False


if __name__ == "__main__":
    projection()



from gantry.dataset.main import (
    _init,
    create_dataset,
    delete_dataset,
    get_dataset,
    list_dataset_versions,
    list_datasets,
    set_working_directory,
)


from gantry.dataset.gantry_dataset import GantryDataset

__all__ = [
    "_init",
    "GantryDataset",
    "create_dataset",
    "get_dataset",
    "list_dataset_versions",
    "list_datasets",
    "delete_dataset",
    "set_working_directory",
]



import json
import logging
import time
from queue import Empty, Queue
from threading import Thread
from typing import List, Optional

from gantry.serializers import EventEncoder

logger = logging.getLogger(__name__)


BATCH_SIZE_RECORD_LIMIT = 500  # 500 records / request.
BATCH_SIZE_LIMIT_BYTES = 5243000  # 5MiB / request
SINGLE_EVENT_LIMIT_BYTES = 1049000  # 1MiB / record
TIMEOUT_SECS = 1  # Max block time for fetching from queue
FLUSH_AFTER_SECS = 1  # Max time to hold data in cache


class _BatchIterBuilder:
    def __init__(
        self,
        queue: Queue,
        batch_size_record_limit: int = BATCH_SIZE_RECORD_LIMIT,
        single_event_limit_bytes: int = SINGLE_EVENT_LIMIT_BYTES,
        batch_size_limit_bytes: int = BATCH_SIZE_LIMIT_BYTES,
        timeout_secs: int = TIMEOUT_SECS,
        flush_after_secs: int = FLUSH_AFTER_SECS,
    ):
        self.queue = queue

        # Serves as cache. Needs to be flushed.
        self.current_json_dumped_batch: List[str] = []
        self.current_json_dumped_batch_size_bytes: int = 0

        self.batch_size_record_limit = batch_size_record_limit
        self.single_event_limit_bytes = single_event_limit_bytes
        self.batch_size_limit_bytes = batch_size_limit_bytes
        self.timeout_secs = timeout_secs
        self.flush_after_secs = flush_after_secs

        self._last_flushed: Optional[float] = None

    def empty(self) -> bool:
        return self.queue.empty() and len(self.current_json_dumped_batch) == 0

    def __iter__(self):
        return self

    def _store_event(self, dumped_json_event):
        self.current_json_dumped_batch.append(dumped_json_event)
        self.current_json_dumped_batch_size_bytes += len(dumped_json_event)

    def _flush(self):
        ret = self.current_json_dumped_batch
        self.current_json_dumped_batch = []
        self.current_json_dumped_batch_size_bytes = 0
        return ret

    def _reset(self, first_event):
        self.current_json_dumped_batch = [first_event]
        self.current_json_dumped_batch_size_bytes = len(first_event)

    def _flush_and_reset(self, dumped_json_event: str):
        # Mark this event as done to enable workers to join
        self.queue.task_done()
        ret = self._flush()
        self._reset(dumped_json_event)
        self._last_flushed = time.time()
        return ret

    def __next__(self):
        """
        If next(it) returns an empty batch, doesn't mean there are no more
        events, as they could be stored in the cache.
        """
        try:
            event = self.queue.get(block=True, timeout=self.timeout_secs)
            dumped_json_event = json.dumps(event, cls=EventEncoder).encode("utf8")

            if len(dumped_json_event) > self.single_event_limit_bytes:
                logger.error(
                    "Dropping record as it is larger than the maximum size allowed (%s bytes)",
                    self.single_event_limit_bytes,
                )
                # Mark this event as done to enable workers to join
                self.queue.task_done()
                return []

            if len(self.current_json_dumped_batch) >= self.batch_size_record_limit:
                logger.debug("Reached batch event size limit. Flushing.")
                return self._flush_and_reset(dumped_json_event)

            if (
                len(dumped_json_event) + self.current_json_dumped_batch_size_bytes
                > self.batch_size_limit_bytes
            ):
                logger.debug("Reached batch limit in bytes. Flushing.")
                return self._flush_and_reset(dumped_json_event)

            if (
                self._last_flushed is None
                or time.time() - self._last_flushed > self.flush_after_secs
            ):
                logger.debug("Reached cache holding limit. Flushing.")
                return self._flush_and_reset(dumped_json_event)

            self._store_event(dumped_json_event)
            # Mark this event as done to enable workers to join
            self.queue.task_done()
            return []

        except Empty:
            logger.debug("No items to consume from queue. Flushing.")
            return self._flush()


class BatchConsumer(Thread):
    def __init__(self, queue, func, batch_iter=_BatchIterBuilder):
        # Daemon threads don't block program exit
        Thread.__init__(self, daemon=True)
        self.queue = queue
        self.func = func

        # It's important to set running in the constructor: if we are asked to
        # pause immediately after construction, we might set running to True in
        # run() *after* we set it to False in pause... and keep running
        # forever.
        self.running = True

        # Should only be provided for testing purposes
        self._batch_iter = batch_iter

    def pause(self):
        self.running = False

    def consume(self, batch_iter):
        batch = next(batch_iter)
        if batch:
            self.func(batch)

    def run(self):
        batch_iter = iter(self._batch_iter(self.queue))
        while self.running or not batch_iter.empty():
            self.consume(batch_iter)



import mock
import pytest

from gantry.alerts import main
from gantry.alerts.client import GantryAlerts


@pytest.mark.parametrize("backend", ["", "/some/file", "tcp://other/protocol"])
def test_global_init_error(backend):
    with mock.patch.dict("os.environ", {"GANTRY_LOGS_LOCATION": backend}):
        with pytest.raises(ValueError):
            main._init()


@pytest.mark.parametrize("valid_backend", ["https://app.gantry.io", "http://app.gantry.io"])
def test_global_init_valid_logs_location(valid_backend):
    with mock.patch.dict("os.environ", {"GANTRY_LOGS_LOCATION": valid_backend}):
        main._init(api_key="ABCD1234")


def test_no_api_key_provided():
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": ""}):
        with pytest.raises(ValueError):
            main._init(api_key=None)


@mock.patch("gantry.alerts.main.globals")
def test_passed_api_key_overrides_env(mock_globals):
    passed_api_key = "some_key"
    env_api_key = "some_other_key"
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": env_api_key}):
        main._init(api_key=passed_api_key)
        assert mock_globals._Alerts._api_client._api_key == passed_api_key
        assert mock_globals._Alerts._api_client._api_key != env_api_key


@mock.patch("gantry.alerts.main.globals")
def test_global_init_default(mock_globals):
    passed_api_key = "some_key"
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": passed_api_key}):
        main._init()  # api_key is required as param or env var
        assert mock_globals._Alerts._api_client._host == "https://app.gantry.io"
        assert mock_globals._Alerts._api_client._api_key == passed_api_key


@mock.patch("gantry.alerts.main.globals")
def test_global_init(mock_globals):
    main._init(api_key="ABCD1234")
    assert mock_globals._Alerts._api_client._host == "https://app.gantry.io"
    assert mock_globals._Alerts._api_client._api_key == "ABCD1234"


@pytest.mark.parametrize(
    ["method", "params"],
    [
        (
            "get_alerts",
            {
                "application_name": "app name",
                "triggered": False,
                "triggered_from": None,
                "triggered_to": None,
            },
        ),
        (
            "get_alerts",
            {
                "application_name": "app name",
                "triggered": True,
                "triggered_from": "1970-01-01T00:00:00",
                "triggered_to": "2022-10-03T14:21:14.454125",
            },
        ),
        (
            "create_alert",
            {
                "application_name": "gantry-demo-logger-14b3910-dev",
                "alert_name": "sdk alert",
                "alert_checks": {
                    "aggregation": "minimum",
                    "data_node_ids": ["fad635d7-13f1-4ceb-ad10-684cf523181b"],
                    "lower_bound": 1000,
                    "upper_bound": 1001,
                },
                "evaluation_window": "PT5M",
                "evaluation_delay": "PT5M",
                "tags": {"env": "dev"},
            },
        ),
        (
            "update_alert",
            {
                "application_name": "gantry-demo-logger-14b3910-dev",
                "alert_id": "13b7ebc2-2cad-4ebe-88df-33a62847d3b9",
                "alert_name": "sdk alert",
                "alert_checks": {
                    "aggregation": "minimum",
                    "data_node_ids": ["fad635d7-13f1-4ceb-ad10-684cf523181b"],
                    "lower_bound": 1000,
                    "upper_bound": 1001,
                },
                "evaluation_window": "PT5M",
                "evaluation_delay": "PT5M",
                "tags": {"env": "dev"},
            },
        ),
        (
            "create_slack_notification",
            {
                "alert_id": "13b7ebc2-2cad-4ebe-88df-33a62847d3b9",
                "notification_name": "Slack notification 1",
                "slack_webhook_url": "hook",
                "notify_daily": True,
                "daily_notification_time": "17:01:00",
            },
        ),
        (
            "create_slack_notification",
            {
                "alert_id": "13b7ebc2-2cad-4ebe-88df-33a62847d3b9",
                "notification_name": "Slack notification 1",
                "slack_webhook_url": "hook",
                "notify_daily": False,
                "daily_notification_time": None,
            },
        ),
        ("get_slack_notifications", {}),
        ("delete_alert", {"alert_id": "13b7ebc2-2cad-4ebe-88df-33a62847d3b9"}),
        (
            "delete_slack_notification",
            {
                "alert_id": "13b7ebc2-2cad-4ebe-88df-33a62847d3b9",
                "slack_notification_id": "a21c4157-7bee-4701-9153-d5d8af2a363b",
            },
        ),
    ],
)
@mock.patch("gantry.alerts.main.validate_init")
@mock.patch("gantry.alerts.main.globals._Alerts", spec=GantryAlerts)
def test_alert_methods(mock_query, mock_validate_init, method, params):
    getattr(mock_query, method).return_value = "foo"
    assert getattr(main, method)(**params) == "foo"
    mock_validate_init.assert_called_once()
    getattr(mock_query, method).assert_called_once_with(**params)



import os
from typing import Optional

from gantry.api_client import APIClient
from gantry.automations import globals
from gantry.const import PROD_API_URL


def _init(
    api_key: Optional[str] = None,
):
    backend = os.environ.get("GANTRY_LOGS_LOCATION", PROD_API_URL)

    api_key = os.environ.get("GANTRY_API_KEY") if api_key is None else api_key

    if not api_key:
        raise ValueError(
            """
            No API key provided. Please pass the api_key parameter or set the GANTRY_API_KEY
            environment variable.
            """
        )

    api_client = APIClient(backend, api_key)
    # instantiated curators need a client, too
    globals._API_CLIENT = api_client  # type: ignore



import logging
import uuid

import pytest
import responses
from responses import matchers

from gantry.applications.llm import (
    CompletionAPILogStore,
    CompletionApplication,
    OpenAIParams,
)
from gantry.logger.stores import BaseLogStore

from ..conftest import TestingBatchConsumer
from .conftest import HOST


class TestingCompletionLogStore(BaseLogStore):
    def __init__(self, *args, **kwargs):
        self._store = []

    def log(self, application: str, event: dict) -> None:
        self._store.append((application, event))


@pytest.fixture
def test_llm_application(test_api_client):
    return CompletionApplication(
        name="test_llm_application",
        api_client=test_api_client,
        id=uuid.uuid4(),
        log_store_factory=TestingCompletionLogStore,
    )


def test_get_llm_application(test_client):
    with responses.RequestsMock() as resp:
        test_id = uuid.uuid4()
        application_name = "test_application_name"
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application_name}",
            json={
                "response": "ok",
                "data": {
                    "versions": [
                        {
                            "internal_version": "1",
                            "version": "first model",
                            "application_id": str(test_id),
                        },
                    ],
                    "func_name": application_name,
                },
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application_name}/schemas",
            json={
                "response": "ok",
                "data": {
                    "model_type": "Completion",
                },
            },
        )
        test_application = test_client.get_application(application_name)
        assert test_application._name == application_name
        assert test_application._id == test_id

        assert isinstance(test_application, CompletionApplication)


@pytest.mark.parametrize(
    ("version_kwargs", "configuration_data"),
    [
        (
            {"prompt_template": "test prompt template", "model": "text-davinci-001"},
            {
                "prompt": "test prompt template",
                "prompt_inputs": [],
                "model": "text-davinci-001",
                "params": OpenAIParams().dict(),
            },
        ),
        (
            {"prompt_template": "test prompt {{template}}", "model": "text-davinci-001"},
            {
                "prompt": "test prompt {{template}}",
                "prompt_inputs": ["template"],
                "model": "text-davinci-001",
                "params": OpenAIParams().dict(),
            },
        ),
        (
            {
                "prompt_template": "x",
                "model": "text-davinci-001",
                "model_params": {"temperature": 0.5, "max_tokens": 10},
            },
            {
                "prompt": "x",
                "prompt_inputs": [],
                "model": "text-davinci-001",
                "params": OpenAIParams(temperature=0.5, max_tokens=10).dict(),
            },
        ),
        (
            {"prompt_template": "x", "model": "text-davinci-001", "prompt_inputs": ["test"]},
            {
                "prompt": "x",
                "prompt_inputs": ["test"],
                "model": "text-davinci-001",
                "params": OpenAIParams().dict(),
            },
        ),
    ],
)
def test_create_version(
    test_llm_application: CompletionApplication, version_kwargs: dict, configuration_data: dict
):
    version_id = uuid.uuid4()
    description = "test description"
    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            (
                f"{HOST}/api/v1/cms/{test_llm_application._name}/configurations/"
                f"{test_llm_application._configuration_name}/versions"
            ),
            json={
                "response": "ok",
                "data": {
                    "id": str(version_id),
                    "configuration_id": str(uuid.uuid4()),
                    "version": 1,
                    "checksum": "1234",
                    "s3_paths": ["s3://test_bucket/test_path"],
                    "notes": description,
                    "created_at": "2020-01-01T00:00:00.000000Z",
                },
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "configuration_data": configuration_data,
                        "description": description,
                    },
                )
            ],
        )

        version = test_llm_application.create_version(
            description=description,
            **version_kwargs,
        )

    assert version.version_id == version_id
    assert version.config == configuration_data


def test_create_version_invalid_model(test_llm_application: CompletionApplication):
    with pytest.raises(ValueError):
        test_llm_application.create_version(
            description="test description",
            prompt_template="test prompt template",
            model="invalid model",
        )


@pytest.mark.parametrize(
    "params",
    [
        {"temperature": "hot"},
    ],
)
def test_create_version_invalid_params(test_llm_application: CompletionApplication, params):
    with pytest.raises(ValueError):
        test_llm_application.create_version(
            description="test description",
            prompt_template="test prompt template",
            model="text-davinci-001",
            model_params=params,
        )


@pytest.mark.parametrize("version", ["latest", "prod", "test"])
def test_get_version(test_llm_application: CompletionApplication, version: str):
    version_id = uuid.uuid4()
    configuration_id = uuid.uuid4()
    configuration_data = {"test": "test"}
    description = "test description"
    version_data = {
        "id": str(version_id),
        "configuration_id": str(configuration_id),
        "version": 1,
        "checksum": "1234",
        "s3_paths": ["s3://test_bucket/test_path"],
        "notes": description,
        "created_at": "2020-01-01T00:00:00.000000Z",
    }

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            (
                f"{HOST}/api/v1/cms/{test_llm_application._name}/configurations/"
                f"{test_llm_application._configuration_name}/versions/{version_id}/data"
            ),
            json={"response": "ok", "data": [configuration_data]},
            headers={"Content-Type": "application/json"},
        )

        if version == "latest":
            resp.add(
                resp.GET,
                (
                    f"{HOST}/api/v1/cms/{test_llm_application._name}/configurations/"
                    f"{test_llm_application._configuration_name}/versions"
                ),
                json={"response": "ok", "data": [version_data, {"fake": "data"}]},
                headers={"Content-Type": "application/json"},
            )
        else:
            resp.add(
                resp.GET,
                (
                    f"{HOST}/api/v1/cms/{test_llm_application._name}/configurations/"
                    f"{test_llm_application._configuration_name}/releases?latest=True&env={version}"
                ),
                json={
                    "response": "ok",
                    "data": [
                        {
                            "id": str(uuid.uuid4()),
                            "configuration_id": str(configuration_id),
                            "description": "test release",
                            "version": version_data,
                            "env": None,
                        }
                    ],
                },
                headers={"Content-Type": "application/json"},
            )

        test_version = test_llm_application.get_version(version)  # type: ignore

    assert test_version is not None
    assert test_version.version_id == version_id
    assert test_version.config == configuration_data


@pytest.mark.parametrize("version", ["latest", "prod", "test"])
def test_get_version_not_found(test_llm_application: CompletionApplication, version: str, caplog):
    with responses.RequestsMock() as resp:
        if version == "latest":
            resp.add(
                resp.GET,
                (
                    f"{HOST}/api/v1/cms/{test_llm_application._name}/configurations/"
                    f"{test_llm_application._configuration_name}/versions"
                ),
                json={"response": "ok", "data": []},
                headers={"Content-Type": "application/json"},
            )
        else:
            resp.add(
                resp.GET,
                (
                    f"{HOST}/api/v1/cms/{test_llm_application._name}/configurations/"
                    f"{test_llm_application._configuration_name}/releases?latest=True&env={version}"
                ),
                json={"response": "ok", "data": []},
                headers={"Content-Type": "application/json"},
            )

        with caplog.at_level(logging.WARNING):
            test_version = test_llm_application.get_version(version)  # type: ignore

    assert "No version found" in caplog.text
    assert test_version is None


def test_log_completion(test_llm_application: CompletionApplication, caplog):
    test_llm_application.log_completion(
        open_ai_api_request={"request": "foo"},
        open_ai_api_response={"response": "bar"},
        request_attributes={"foo": "bar"},
        response_attributes={"bar": "baz"},
        feedback={"foo": "baz"},
        selected_choice_index=1,
        session_id="12345",
        tags={"a": "b"},
        version=100,
    )

    assert test_llm_application._log_store._store == [
        (
            "test_llm_application",
            {
                "application": "test_llm_application",
                "open_ai_api_request": {"request": "foo"},
                "open_ai_api_response": {"response": "bar"},
                "request_attributes": {"foo": "bar"},
                "response_attributes": {"bar": "baz"},
                "feedback": {"foo": "baz"},
                "selected_choice_index": 1,
                "session_id": "12345",
                "tags": {"a": "b", "version_number": 100},
            },
        ),
    ]


def test_completion_api_log_store_log_batch(test_api_client):
    log_store = CompletionAPILogStore(test_api_client, consumer_factory=TestingBatchConsumer)
    event = {"foo": "bar"}
    with pytest.raises(NotImplementedError):
        log_store.log_batch("test_llm_application", [event])


def test_completion_api_log_store_log(test_api_client):
    log_store = CompletionAPILogStore(test_api_client, consumer_factory=TestingBatchConsumer)
    event = {"foo": "bar"}
    log_store.log("test_llm_application", event)

    assert log_store.queue.get() == {"foo": "bar"}
    assert log_store.queue.empty()
    assert log_store.consumers[0].running
    assert log_store.consumers[0].func == log_store.consumer_func


def test_completion_api_log_store_consumer_func(test_api_client):
    log_store = CompletionAPILogStore(test_api_client, consumer_factory=TestingBatchConsumer)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/logging/completions",
            json={"response": "ok"},
            headers={"Content-Type": "application/json"},
            match=[matchers.json_params_matcher({"foo": "bar"})],
        )
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/logging/completions",
            json={"response": "ok"},
            headers={"Content-Type": "application/json"},
            match=[matchers.json_params_matcher({"bar": "baz"})],
        )

        log_store.consumer_func([b'{"foo": "bar"}', b'{"bar": "baz"}'])


def test_completion_api_log_store_send_completion_event(test_api_client):
    log_store = CompletionAPILogStore(test_api_client, consumer_factory=TestingBatchConsumer)
    event = {"foo": "bar"}
    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/logging/completions",
            json={"response": "ok"},
            headers={"Content-Type": "application/json"},
            match=[matchers.json_params_matcher({"foo": "bar"})],
        )

        log_store.send_completion_event(event)


def test_deploy_version(test_llm_application: CompletionApplication):
    version_id = uuid.uuid4()
    env = "foobar"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.PATCH,
            (
                f"{HOST}/api/v1/cms/test_llm_application/configurations/"
                "default_test_llm_application_configuration/releases"
            ),
            json={"response": "ok"},
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher(
                    {
                        "version_id": str(version_id),
                        "env": env,
                        "description": "Deployment from SDK",
                    }
                )
            ],
        )

        test_llm_application.deploy_version(version_id, env)



import logging
from typing import Optional

from gantry.api_client import APIClient
from gantry.automations import globals
from gantry.automations.actions import Action, SendSlackMessage
from gantry.automations.curators import Curator
from gantry.automations.curators.main import get_curator
from gantry.automations.triggers import AggregationTrigger, IntervalTrigger, Trigger

logger = logging.getLogger(__name__)


class Automation:
    def __init__(
        self,
        name: str,
        trigger: Trigger,
        action: Action,
        api_client: Optional[APIClient] = None,
        application: Optional[str] = None,
    ):
        """
        Automation is the process of automatically executing tasks.
        It consists of 2 main components: Trigger and Action.

        Trigger is the condition that, when met, cause the action to be executed.

        Currently, we are supporting:

        - **IntervalTrigger**: based on time intervals & delays
        - **AggregationTrigger**: same idea of Alert on the UI. It tracks if an aggregation
            is outside of the range for the evaluation window and triggers when conditions are met.
        - **SendSlackMessage** (Action): This action sets up a notification setting that will send
        users a slack notification when executed by the trigger.
        - **Curator** (Action): This action creates a new curator.
        - Note:
            - `AggregationTrigger` can only be paired with `SendSlackMessage`
            - `IntervalTrigger` can only be paired with `Curator`

        To setup and start an automation process, there are a few steps:
        - Define the trigger
            .. code-block:: python

                from gantry.automations.triggers import AggregationTrigger
                from gantry.query.time_window import RelativeTimeWindow

                alert_trigger = AggregationTrigger(
                    name = "alert-trigger",
                    aggregation = "maximum",
                    fields = ["inputs.speed"],
                    lower_bound = 1,
                    upper_bound = 5,
                    evaluation_window=RelativeTimeWindow(
                        window_length=datetime.timedelta(minutes=10)
                    ),
                    tags = None
                )

        - Define the action
            .. code-block:: python

                from gantry.automations.actions import SendSlackMessage
                # Set up action to send slack message to the specified webhook.
                slack_action = SendSlackMessage(
                    name = "demo-slack",
                    webhook_url="SLACK_WEBHOOK_URL"
                )

        - Define the automation and put trigger & action together
            .. code-block:: python

                from gantry.automations.automations import Automation
                # Define automation object and put trigger and action together.
                automation_alert = Automation(
                    name = "alert-automation",
                    trigger=alert_trigger,
                    action=slack_action
                )
        - Add the automation to an application
            .. code-block:: python

                app.add_automation(automation_alert)

        """
        self.name = name
        self.trigger = trigger
        self.action = action
        self.application = application
        self._api_client = globals._API_CLIENT if api_client is None else api_client  # type: ignore

    def add_to_application(self, application_name):
        """
        Add automation to an application.
        This method shouldn't be called directly.
        It will be called once you call `application.add_automation()`.
        """
        self.application = application_name
        self.trigger.add_to_application(application_name)
        logger.info("Automation has been added to application.")
        self.prep()
        self.start()

    def remove_from_application(self):
        """
        Remove automation from an application.
        This method shouldn't be called directly.
        It will be called once you stop the automation process.
        """
        self.application = None
        self.trigger.remove_from_application()
        logger.info("Automation has been removed from application.")

    @classmethod
    def from_dict(cls, data):
        trigger = Trigger()
        if data["trigger"]["type"] == "AggregationTrigger":
            trigger = AggregationTrigger.from_dict(data["trigger"]["content"])
        elif data["trigger"]["type"] == "IntervalTrigger":
            trigger = IntervalTrigger.from_dict(data["trigger"]["content"])

        action = Action()
        if data["action"]["type"] == "SendSlackMessage":
            action = SendSlackMessage.from_dict(data["action"]["content"])
        elif data["action"]["type"] == "curator":
            action = get_curator(data["action"]["content"]["curator_name"])
        return cls(
            name=data["name"],
            application=data["application"],
            trigger=trigger,
            action=action,
        )

    def to_dict(self):
        action_dict = None
        if isinstance(self.action, Curator):
            action_dict = {
                "type": "curator",
                "content": {
                    "curator_name": self.action._name,
                    "curated_dataset": self.action._curated_dataset_name,
                },
            }
        else:
            action_dict = self.action.to_dict()
        return {
            "name": self.name,
            "application_name": self.application,
            "trigger": self.trigger.to_dict(),
            "action": action_dict,
        }

    def prep(self):
        if isinstance(self.trigger, IntervalTrigger) and isinstance(self.action, Curator):
            if self.trigger.start_on:
                self.action.update_curation_start_on(self.trigger.start_on)
            if self.trigger.interval:
                self.action.update_curation_interval(self.trigger.interval)
            if self.trigger.delay:
                self.action.update_curation_delay(self.trigger.delay)
        if isinstance(self.trigger, AggregationTrigger):
            alert_id = self.trigger.generate_alert()
            self.action.update_trigger(alert_id)

    def start(self):
        self.action.start()
        data = self.to_dict()
        self._api_client.request("POST", "/api/v1/automations", json=data, raise_for_status=True)
        logger.info("Automation has started.")

    def stop(self):
        self._api_client.request(
            "DELETE",
            f"/api/v1/automations/{self.name}",
            params={"application_name": self.application},
            raise_for_status=True,
        )
        self.remove_from_application()
        self.action.stop()
        logger.info("Automation has stopped.")



from gantry.query import globals
from gantry.query.client import GantryQuery
from gantry.query.globals import _query_alias, validate_init


@_query_alias
def d1(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.distance.d1(*args, **kwargs)


@_query_alias
def dinf(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.distance.dinf(*args, **kwargs)


@_query_alias
def ks(*args, **kwargs) -> float:
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.distance.ks(*args, **kwargs)


@_query_alias
def kl(*args, **kwargs) -> float:
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.distance.kl(*args, **kwargs)
